"""
Module with handy utilities for plotting genomic signal
"""
from itertools import groupby
import matplotlib
from matplotlib import pyplot as plt
from matplotlib import mlab
import numpy as np
from matplotlib.ticker import MaxNLocator
from mpl_toolkits.axes_grid1 import make_axes_locatable
from matplotlib import gridspec
import colormap_adjust
from scipy import stats


def ci_plot(x, arr, conf=0.95, ax=None, line_kwargs=None, fill_kwargs=None):
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111)

    line_kwargs = line_kwargs or {}
    fill_kwargs = fill_kwargs or {}

    m, lo, hi = ci(arr, conf)
    ax.plot(x, m, **line_kwargs)
    ax.fill_between(x, lo, hi, **fill_kwargs)
    return ax


def imshow(arr, x=None, ax=None, vmin=None, vmax=None, percentile=True,
           strip=False, features=None, conf=0.95, line_kwargs=None,
           sort_by=None, fill_kwargs=None, figsize=(5, 12),
           width_ratios=(4, 1), height_ratios=(4, 1),
           subplot_params=dict(wspace=0.1, hspace=0.1), imshow_kwargs=None):
    """
    Parameters
    ----------
    arr : array-like

    x : 1D array
        X values to use.  If None, use range(arr.shape[1])

    ax : matplotlib.Axes
        If not None, then only plot the array on the provided axes.  This will
        ignore any additional arguments provided that apply to figure-level
        configuration or to the average line plot.  For example, `figsize`,
        `width_ratios`, `height_ratios`, `subplot_params`, `line_kwargs`, and
        `fill_kwargs` will be ignored.

    vmin, vmax : float

    percentile : bool
        If True, then treat values for `vmin` and `vmax` as percentiles rather
        than absolute values.

    strip : bool
        Include a strip plot alongside the array

    features : pybedtools.BedTool or string filename
        Features used to construct the array

    sort_by : array-like
        Use the provided array to sort the array (e.g., expression).  This
        array is argsorted to get the proper order.

    line_kwargs, fill_kwargs : dict
        Passed directly to `ci_plot`.

    figsize : tuple
        (Width, height) of the figure to create.
    """
    if ax is None:
        fig = new_shell(
            figsize=figsize,
            strip=strip,
            subplot_params=subplot_params,
            width_ratios=width_ratios,
            height_ratios=height_ratios)

    if x is None:
        x = np.arange(arr.shape[1])

    if percentile:
        if vmin is None:
            vmin = arr.min()
        else:
            vmin = mlab.prctile(arr.ravel(), vmin)
        if vmax is None:
            vmax = arr.max()
        else:
            vmax = mlab.prctile(arr.ravel(), vmax)
    else:
        if vmin is None:
            vmin = arr.min()
        if vmax is None:
            vmax = arr.max()

    if imshow_kwargs is None:
        imshow_kwargs = {}

    cmap = colormap_adjust.smart_colormap(vmin, vmax)
    if sort_by is not None:
        ind = np.argsort(sort_by)
    else:
        ind = np.arange(arr.shape[0])

    if ax is None:
        array_ax = fig.array_axes
    else:
        array_ax = ax

    mappable = array_ax.imshow(
        arr[ind, :],
        aspect='auto',
        cmap=cmap,
        vmin=vmin,
        vmax=vmax,
        origin='lower',
        extent=(x.min(), x.max(), 0, arr.shape[0]),
        **imshow_kwargs
    )
    if ax is None:
        plt.colorbar(mappable, fig.cax)
        ci_plot(
            x,
            arr,
            ax=fig.line_axes,
            line_kwargs=line_kwargs,
            fill_kwargs=fill_kwargs,
        )

        return fig
    else:
        return ax.figure


def calculate_limits(array_dict, method='global', percentiles=None, limit=()):
    """
    Calculate limits for a group of arrays in a flexible manner.

    Returns a dictionary of calculated (vmin, vmax), with the same keys as
    `array_dict`.

    Useful for plotting heatmaps of multiple datasets, and the vmin/vmax values
    of the colormaps need to be matched across all (or a subset) of heatmaps.

    Parameters
    ----------
    array_dict : dict of np.arrays

    method : {'global', 'independent', callable}
        If method="global", then use the global min/max values across all
        arrays in array_dict.  If method="independent", then each array will
        have its own min/max calcuated.  If a callable, then it will be used to
        group the keys of `array_dict`, and each group will have its own
        group-wise min/max calculated.

    limit: tuple, optional
        Tuple of 2 scalars passed directly to matplotlib.mlab.prctile to
        limit the calculation of the percentile.

    percentiles : None or list
        If not None, a list of (lower, upper) percentiles in the range [0,100].
    """
    if percentiles is not None:
        for percentile in percentiles:
            if not 0 <= percentile <= 100:
                raise ValueError("percentile (%s) not between [0, 100]")

    if method == 'global':
        all_arrays = np.concatenate(
            [i.ravel() for i in array_dict.itervalues()]
        )
        if percentiles:
            vmin = mlab.prctile(
                all_arrays, percentiles[0], limit=limit)
            vmax = mlab.prctile(
                all_arrays, percentiles[1], limit=limit)

        else:
            vmin = all_arrays.min()
            vmax = all_arrays.max()
        d = dict([(i, (vmin, vmax)) for i in array_dict.keys()])

    elif method == 'independent':
        d = {}
        for k, v in array_dict.iteritems():
            d[k] = (v.min(), v.max())

    elif hasattr(method, '__call__'):
        d = {}
        sorted_keys = sorted(array_dict.keys(), key=method)
        for group, keys in groupby(sorted_keys, method):
            keys = list(keys)
            all_arrays = np.concatenate([array_dict[i] for i in keys])
            if percentiles:
                vmin = mlab.prctile(
                    all_arrays, percentiles[0], limit=limit)
                vmax = mlab.prctile(
                    all_arrays, percentiles[1], limit=limit)
            else:
                vmin = all_arrays.min()
                vmax = all_arrays.max()
            for key in keys:
                d[key] = (vmin, vmax)
    return d


def ci(arr, conf=0.95):
    """
    Column-wise confidence interval.

    Parameters
    ----------
    arr : array-like

    conf : float
        Confidence interval

    Returns
    -------
    m : array
        column-wise mean
    lower : array
        lower column-wise confidence bound
    upper : array
        upper column-wise confidence bound
    """
    m = arr.mean(axis=0)
    n = len(arr)
    se = arr.std(axis=0) / np.sqrt(n)
    h = se * stats.t._ppf((1 + conf) / 2., n - 1)
    return m, m - h, m + h


def nice_log(x):
    """
    Uses a log scale but with negative numbers.

    :param x: NumPy array
    """
    neg = x < 0
    xi = np.log2(np.abs(x) + 1)
    xi[neg] = -xi[neg]
    return xi


def tip_zscores(a):
    """
    Calculates the "target identification from profiles" (TIP) zscores
    from Cheng et al. 2001, Bioinformatics 27(23):3221-3227.

    :param a: NumPy array, where each row is the signal for a feature.
    """
    weighted = a * a.mean(axis=0)
    scores = weighted.sum(axis=1)
    zscores = (scores - scores.mean()) / scores.std()
    return zscores


def fdrcorrection(pvals, alpha=0.05, method='indep'):
    '''
    NOTE: This function was copied from
    statsmodels.sandbox.stats.multicomp.fdrcorrection0, from statsmodels
    version 0.5.0.

    This is to avoid requiring all of statsmodels to be a dependency for
    metaseq, just for this function.




    pvalue correction for false discovery rate

    This covers Benjamini/Hochberg for independent or positively correlated and
    Benjamini/Yekutieli for general or negatively correlated tests. Both are
    available in the function multipletests, as method=`fdr_bh`, resp. `fdr_by`.

    Parameters
    ----------
    pvals : array_like
        set of p-values of the individual tests.
    alpha : float
        error rate
    method : {'indep', 'negcorr')

    Returns
    -------
    rejected : array, bool
        True if a hypothesis is rejected, False if not
    pvalue-corrected : array
        pvalues adjusted for multiple hypothesis testing to limit FDR

    Notes
    -----

    If there is prior information on the fraction of true hypothesis, then alpha
    should be set to alpha * m/m_0 where m is the number of tests,
    given by the p-values, and m_0 is an estimate of the true hypothesis.
    (see Benjamini, Krieger and Yekuteli)

    The two-step method of Benjamini, Krieger and Yekutiel that estimates the number
    of false hypotheses will be available (soon).

    Method names can be abbreviated to first letter, 'i' or 'p' for fdr_bh and 'n' for
    fdr_by.

    '''
    pvals = np.asarray(pvals)

    pvals_sortind = np.argsort(pvals)
    pvals_sorted = pvals[pvals_sortind]
    sortrevind = pvals_sortind.argsort()

    if method in ['i', 'indep', 'p', 'poscorr']:
        ecdffactor = _ecdf(pvals_sorted)
    elif method in ['n', 'negcorr']:
        cm = np.sum(1./np.arange(1, len(pvals_sorted)+1))   #corrected this
        ecdffactor = _ecdf(pvals_sorted) / cm
##    elif method in ['n', 'negcorr']:
##        cm = np.sum(np.arange(len(pvals)))
##        ecdffactor = ecdf(pvals_sorted)/cm
    else:
        raise ValueError('only indep and necorr implemented')
    reject = pvals_sorted <= ecdffactor*alpha
    if reject.any():
        rejectmax = max(np.nonzero(reject)[0])
        reject[:rejectmax] = True

    pvals_corrected_raw = pvals_sorted / ecdffactor
    pvals_corrected = np.minimum.accumulate(pvals_corrected_raw[::-1])[::-1]
    pvals_corrected[pvals_corrected>1] = 1
    return reject[sortrevind], pvals_corrected[sortrevind]


def tip_fdr(a, alpha=0.05):
    """
    Returns adjusted TIP p-values for a particular `alpha`.

    (see :func:`tip_zscores` for more info)

    :param a: NumPy array, where each row is the signal for a feature
    :param alpha: False discovery rate

    """
    zscores = tip_zscores(a)
    pvals = stats.norm.pdf(zscores)
    rejected, fdrs = fdrcorrection(pvals)
    return fdrs


def prepare_logged(x, y):
    """
    Transform `x` and `y` to a log scale while dealing with zeros.

    This function scales `x` and `y` such that the points that are zero in one
    array are set to the min of the other array.

    When plotting expression data, frequently one sample will have reads in
    a particular feature but the other sample will not.  Expression data also
    tends to look better on a log scale, but log(0) is undefined and therefore
    cannot be shown on a plot.  This function allows these points to be shown,
    piled up along one side of the plot.

    :param x,y: NumPy arrays
    """
    xi = np.log2(x)
    yi = np.log2(y)

    xv = np.isfinite(xi)
    yv = np.isfinite(yi)

    global_min = min(xi[xv].min(), yi[yv].min())
    global_max = max(xi[xv].max(), yi[yv].max())

    xi[~xv] = global_min
    yi[~yv] = global_min

    return xi, yi


def new_shell(figsize=(5, 12), strip=False, height_ratios=(4, 1),
              width_ratios=(4, 1), subplot_params=None):
    if subplot_params is None:
        subplot_params = {}
    fig = plt.figure(figsize=figsize)
    gs = gridspec.GridSpec(
        len(height_ratios),
        len(width_ratios),
        height_ratios=height_ratios,
        width_ratios=width_ratios,
        **subplot_params)
    fig.array_axes = plt.subplot(gs[0, 0])
    if strip:
        fig.strip_axes = plt.subplot(gs[0, 1])
    fig.line_axes = plt.subplot(gs[1, 0])
    fig.cax = plt.subplot(gs[1, 1])
    fig.gs = gs
    return fig


def matrix_and_line_shell(figsize=(5, 12), strip=False):
    """
    Helper function to construct an empty figure that has space for a matrix,
    a summary line plot directly below it, a colorbar axis, and an optional
    "strip" axis that parallels the matrix (and shares its y-axis) where data
    can be added to create callbacks.

    Returns a tuple of (fig, matrix_ax, line_ax, strip_ax, colorbar_ax) that
    can then be used to plot upon.

    :param figsize: Tuple of (width, height), in inches, for the figure to be
    :param strip: If `strip` is False, then the returned `strip_ax` will be
        None and no strip axes will be created.
    """
    fig = plt.figure(figsize=figsize)

    # Constants to keep track
    if strip:
        STRIP_COLS = 1
    else:
        STRIP_COLS = 0
    ROWS = 4
    COLS = 8 + STRIP_COLS
    MAT_COLS = 7
    MAT_ROWS = 3
    LINE_ROWS = ROWS - MAT_ROWS

    mat_ax = plt.subplot2grid(
        shape=(ROWS, COLS),
        loc=(0, STRIP_COLS),
        rowspan=MAT_ROWS,
        colspan=MAT_COLS,
    )

    line_ax = plt.subplot2grid(
        shape=(ROWS, COLS),
        loc=(MAT_ROWS, STRIP_COLS),
        rowspan=LINE_ROWS,
        colspan=MAT_COLS,
        sharex=mat_ax)

    if strip:
        strip_ax = plt.subplot2grid(
            shape=(ROWS, COLS),
            loc=(0, 0),
            rowspan=MAT_ROWS,
            colspan=STRIP_COLS,
            sharey=mat_ax,
        )
    else:
        strip_ax = None

    cax = plt.subplot2grid(
        shape=(ROWS, COLS),
        loc=(ROWS - MAT_ROWS, MAT_COLS + STRIP_COLS),
        rowspan=1,
        colspan=1,
    )

    fig.subplots_adjust(hspace=0.1, wspace=0.2, right=0.88, left=0.23)
    return fig, mat_ax, line_ax, strip_ax, cax


def clustered_sortind(x, k=10, scorefunc=None):
    """
    Uses MiniBatch k-means clustering to cluster matrix into groups.

    Each cluster of rows is then sorted by `scorefunc` -- by default, the max
    peak height when all rows in a cluster are averaged, or
    cluster.mean(axis=0).max().

    Returns the index that will sort the rows of `x` and a list of "breaks".
    `breaks` is essentially a cumulative row count for each cluster boundary.
    In other words, after plotting the array you can use axhline on each
    "break" to plot the cluster boundary.

    If `k` is a list or tuple, iteratively try each one and select the best
    with the lowest mean distance from cluster centers.

    :param x: Matrix whose rows are to be clustered
    :param k: Number of clusters to create or a list of potential clusters; the
        optimum will be chosen from the list
    :param scorefunc: Optional function for sorting rows within clusters.  Must
        accept a single argument of a NumPy array.
    """
    try:
        from sklearn.cluster import MiniBatchKMeans
    except ImportError:
        raise ImportError('please install scikits.learn for '
                          'clustering.')

    # If integer, do it once and we're done
    if isinstance(k, int):
        best_k = k

    else:
        mean_dists = {}
        for _k in k:
            mbk = MiniBatchKMeans(init='k-means++', n_clusters=_k)
            mbk.fit(x)
            mean_dists[_k] = mbk.transform(x).mean()
        best_k = sorted(mean_dists.items(), key=lambda x: x[1])[-1][0]

    mbk = MiniBatchKMeans(init='k-means++', n_clusters=best_k)
    mbk.fit(x)
    k = best_k
    labels = mbk.labels_
    scores = np.zeros(labels.shape, dtype=float)

    if not scorefunc:
        def scorefunc(x):
            return x.mean(axis=0).max()

    for label in range(k):
        ind = labels == label
        score = scorefunc(x[ind, :])
        scores[ind] = score

    pos = 0
    breaks = []
    ind = np.argsort(scores)
    for k, g in groupby(labels[ind]):
        pos += len(list(g))
        breaks.append(pos)

    return ind, breaks


def new_clustered_sortind(x, k=10, row_key=None, cluster_key=None):
    """
    Uses MiniBatch k-means clustering to cluster matrix into groups.

    Each cluster of rows is then sorted by `scorefunc` -- by default, the max
    peak height when all rows in a cluster are averaged, or
    cluster.mean(axis=0).max().

    Returns the index that will sort the rows of `x` and a list of "breaks".
    `breaks` is essentially a cumulative row count for each cluster boundary.
    In other words, after plotting the array you can use axhline on each
    "break" to plot the cluster boundary.

    If `k` is a list or tuple, iteratively try each one and select the best
    with the lowest mean distance from cluster centers.

    :param x: Matrix whose rows are to be clustered
    :param k: Number of clusters to create or a list of potential clusters; the
        optimum will be chosen from the list
    :param row_key:
        Optional function to act as a sort key for sorting rows within
        clusters.  Signature should be `scorefunc(a)` where `a` is a 1-D NumPy
        array.
    :param cluster_key:
        Optional function for sorting clusters.  Signature is `clusterfunc(a)`
        where `a` is a NumPy array containing all rows of `x` for cluster `i`.
        It must return a single value.
    """
    try:
        from sklearn.cluster import MiniBatchKMeans
    except ImportError:
        raise ImportError('please install scikits.learn for '
                          'clustering.')

    # If integer, do it once and we're done
    if isinstance(k, int):
        best_k = k

    else:
        mean_dists = {}
        for _k in k:
            mbk = MiniBatchKMeans(init='k-means++', n_clusters=_k)
            mbk.fit(x)
            mean_dists[_k] = mbk.transform(x).mean()
        best_k = sorted(mean_dists.items(), key=lambda x: x[1])[-1][0]

    mbk = MiniBatchKMeans(init='k-means++', n_clusters=best_k)
    mbk.fit(x)
    k = best_k
    labels = mbk.labels_
    scores = np.zeros(labels.shape, dtype=float)

    if cluster_key:
        # It's easier for calling code to provide something that operates on
        # a cluster level, but here it's converted to work on a label level
        # that looks in to the array `x`.
        def _cluster_key(i):
            return cluster_key(x[labels == i, :])
        sorted_labels = sorted(range(k), key=_cluster_key)
    else:
        # Otherwise just use them as-is.
        sorted_labels = range(k)

    if row_key:
        # Again, easier to provide a function to operate on a row.  But here we
        # need it to accept an index
        def _row_key(i):
            return row_key(x[i, :])

    final_ind = []
    breaks = []
    pos = 0
    for label in sorted_labels:
        # which rows in `x` have this label
        label_inds = np.nonzero(labels == label)[0]
        if row_key:
            label_sort_ind = sorted(label_inds, key=_row_key)
        else:
            label_sort_ind = label_inds
        for li in label_sort_ind:
            final_ind.append(li)
        pos += len(label_inds)
        breaks.append(pos)

    return np.array(final_ind), np.array(breaks)


def input_ip_plots(iparr, inputarr, diffed, x, sort_ind,
                   prefix=None, limits1=(None, None), limits2=(None, None),
                   hlines=None, vlines=None):

    """
    All-in-one plotting function to make a 5-panel figure.

    Panels are IP, input, and diffed; plus 2 line plots showing averages.

    :param iparr, inputarr: NumPy arrays constructed by a genomic_signal object
    :param diffed: Difference of `iparr` and `inputarr`, but can be some other
                   transformation.
    :param x: Extent to use -- for TSSs, maybe something like
        np.linspace(-1000, 1000, bins), or for just bin IDs, something like
        `np.arange(bins)`.

    :param sort_ind: row order for each of the 3 panels -- usually interesting
        to use `clustered_sortind` or `tip_zscores`

    :param prefix: Used to prefix plot titles with '%(prefix)s IP", etc
    :param limits1: Tuple passed to the Normalize function for IP and input.
    :param limits2: Tuple passed tot he Normalize function for the diffed array
    :param hlines: List of (position, kwarg) tuples for plotting horizontal
        lines.  Kwargs are passed directly to axhline. Useful for delimiting
        clusters, if you used `clustered_sortind` and have both `row_order` and
        `breaks`.
    :param vlines: List of (position, kwargs) tuples.  A vertical line will be
        plotted at each position using kwargs.
    """

    # global min and max
    gmin = min(iparr.min(), inputarr.min())
    gmax = max(iparr.max(), inputarr.max())

    fig = plt.figure(figsize=(10, 10))

    # 3 arrays, 2 line plots, a gene strip, and 2 colorbars.  Plots share the
    # axes that make sense
    #
    # 3 arrays
    ax1 = plt.subplot2grid(
        (9, 9), (0, 0), colspan=3, rowspan=6)
    ax2 = plt.subplot2grid(
        (9, 9), (0, 3), colspan=3, rowspan=6, sharex=ax1, sharey=ax1)
    ax3 = plt.subplot2grid(
        (9, 9), (0, 6), colspan=3, rowspan=6, sharex=ax1, sharey=ax1)

    # 2 line plots
    ax4 = plt.subplot2grid((9, 9), (6, 3), colspan=3, rowspan=3, sharex=ax1)
    ax5 = plt.subplot2grid((9, 9), (6, 6), colspan=3, rowspan=3, sharex=ax1)

    # 2 colorbars
    cax1 = plt.Axes(fig, rect=(0.05, 0.25, 0.25, 0.025))
    cax2 = plt.Axes(fig, rect=(0.05, 0.15, 0.25, 0.025))

    # For nice imshow axes
    extent = (min(x), max(x), 0, diffed.shape[0])

    cm = matplotlib.cm.gist_gray
    cm.set_bad('k')
    cm.set_over('r')
    cm.set_under('b')

    limits1 = list(limits1)
    limits2 = list(limits2)

    all_base = np.column_stack((iparr.ravel(), inputarr.ravel())).ravel()

    if limits1[0] is None:
        limits1[0] = mlab.prctile(
            all_base, 1. / all_base.size)
    if limits1[1] is None:
        limits1[1] = mlab.prctile(
            all_base, 100 - 1. / all_base.size)
    if limits2[0] is None:
        limits2[0] = mlab.prctile(
            diffed.ravel(), 1. / all_base.size)
    if limits2[1] is None:
        limits2[1] = mlab.prctile(
            diffed.ravel(), 100 - 1. / all_base.size)

    del all_base

    imshow_kwargs = dict(
        interpolation='nearest',
        aspect='auto',
        cmap=cm,
        norm=matplotlib.colors.Normalize(*limits1),
        extent=extent,
        origin='lower')

    # modify kwargs for diffed (by changing the normalization)
    diffed_kwargs = imshow_kwargs.copy()
    diffed_kwargs['norm'] = matplotlib.colors.Normalize(*limits2)

    # IP
    mappable1 = ax1.imshow(iparr[sort_ind, :], **imshow_kwargs)

    # input
    mappable2 = ax2.imshow(inputarr[sort_ind, :], **imshow_kwargs)

    # diffed
    mappable3 = ax3.imshow((diffed)[sort_ind, :], **diffed_kwargs)

    # IP and input line plot with vertical line
    ax4.plot(x, inputarr.mean(axis=0), color='k', linestyle='--',
             label='input')
    ax4.plot(x, iparr.mean(axis=0), color='k', label='ip')
    ax4.axvline(0, color='k', linestyle=':')

    # Diffed line plot with vertical line
    ax5.plot(x, diffed.mean(axis=0), 'k', label='enrichment')
    ax5.axvline(0, color='k', linestyle=':')

    # Colorbars
    cbar1 = fig.colorbar(mappable1, cax1, orientation='horizontal')
    cbar2 = fig.colorbar(mappable3, cax2, orientation='horizontal')
    fig.add_axes(cax1)
    fig.add_axes(cax2)

    # labeling...
    ax1.set_ylabel('features')
    plt.setp(ax2.get_yticklabels(), visible=False)
    plt.setp(ax3.get_yticklabels(), visible=False)
    ax4.set_xlabel('bp')
    ax4.set_ylabel('mean reads per million mapped reads')
    ax5.set_xlabel('bp')
    cax1.set_xlabel('Reads per million mapped reads')
    cax2.set_xlabel('Enrichment (RPMMR)')

    if prefix is None:
        prefix = ""
    ax1.set_title('%s IP' % prefix)
    ax2.set_title('%s input' % prefix)
    ax3.set_title('Difference')

    # diffed line plot should have y ax on right
    ax5.yaxis.set_ticks_position('right')
    ax5.yaxis.set_label_position('right')
    ax5.set_ylabel('enriched reads per million mapped reads')

    # Legends
    ax4.legend(loc='best', frameon=False)
    ax5.legend(loc='best', frameon=False)

    # Make sure everybody snaps to xmin/xmax
    for ax in [ax1, ax2, ax3, ax4, ax5]:
        ax.axis(xmin=extent[0], xmax=extent[1])

    if not hlines:
        hlines = []
    if not vlines:
        vlines = []

    for ax in [ax1, ax2, ax3]:
        for pos, kwargs in hlines:
            ax.axhline(pos, **kwargs)
        for pos, kwargs in vlines:
            ax.axvline(pos, **kwargs)

    fig.subplots_adjust(bottom=0.05, top=0.95, hspace=0.75, wspace=0.9)

    return fig


def _updatecopy(orig, update_with, keys=None, override=False):
    """
    Update a copy of dest with source.  If `keys` is a list, then only update
    with those keys.
    """
    d = orig.copy()
    if keys is None:
        keys = update_with.keys()
    for k in keys:
        if k in update_with:
            if k in d and not override:
                continue
            d[k] = update_with[k]
    return d


def _clean(z):
    """
    Return a version of z that only has finite values
    """
    return z[np.isfinite(z)]


class MarginalHistScatter(object):
    def __init__(self, ax, hist_size=0.6, pad=0.05):
        """
        Class to enable incremental appending of scatterplots, each of which
        generate additional marginal histograms.
        """
        self.scatter_ax = ax
        self.fig = ax.figure

        self.divider = make_axes_locatable(self.scatter_ax)

        self.top_hists = []
        self.right_hists = []
        self.hist_size = hist_size
        self.pad = pad
        self.xfirst_ax = None
        self.yfirst_ax = None

        # will hold histogram data
        self.hxs = []
        self.hys = []

    @property
    def xmax(self):
        return self.scatter_ax.dataLim.xmax

    @property
    def ymax(self):
        return self.scatter_ax.dataLim.ymax

    @property
    def xmin(self):
        return self.scatter_ax.dataLim.xmin

    @property
    def ymin(self):
        return self.scatter_ax.dataLim.ymin

    @property
    def limits(self):
        return (self.xmin, self.xmax, self.ymin, self.ymax)

    def append(self, x, y, scatter_kwargs, hist_kwargs=None, xhist_kwargs=None,
               yhist_kwargs=None, num_ticks=3, labels=None, hist_share=False,
               marginal_histograms=True):
        """
        Adds a new scatter to self.scatter_ax as well as marginal histograms
        for the same data, borrowing addtional room from the axes.

        Parameters
        ----------

        x, y : array-like
            Data to be plotted

        scatter_kwargs : dict
            Keyword arguments that are passed directly to scatter().

        hist_kwargs : dict
            Keyword arguments that are passed directly to hist(), for both the
            top and side histograms.

        xhist_kwargs, yhist_kwargs : dict
            Additional, margin-specific kwargs for the x or y histograms
            respectively.  These are used to update `hist_kwargs`

        num_ticks : int
            How many tick marks to use in each histogram's y-axis

        labels : array-like
            Optional NumPy array of labels that will be set on the collection
            so that they can be accessed by a callback function.

        hist_share : bool
            If True, then all histograms will share the same frequency axes.
            Useful for showing relative heights if you don't want to use the
            hist_kwarg `normed=True`

        marginal_histograms : bool
            Set to False in order to disable marginal histograms and just use
            as a normal scatterplot.
        """
        scatter_kwargs = scatter_kwargs or {}
        hist_kwargs = hist_kwargs or {}
        xhist_kwargs = xhist_kwargs or {}
        yhist_kwargs = yhist_kwargs or {}
        yhist_kwargs.update(dict(orientation='horizontal'))

        # Plot the scatter
        self.scatter_ax.scatter(x, y, **scatter_kwargs)
        self.scatter_ax.collections[-1].labels = labels

        if not marginal_histograms:
            return

        xhk = _updatecopy(hist_kwargs, xhist_kwargs)
        yhk = _updatecopy(hist_kwargs, yhist_kwargs)

        axhistx = self.divider.append_axes(
            'top', size=self.hist_size,
            pad=self.pad, sharex=self.scatter_ax, sharey=self.xfirst_ax)

        axhisty = self.divider.append_axes(
            'right', size=self.hist_size,
            pad=self.pad, sharey=self.scatter_ax, sharex=self.yfirst_ax)

        axhistx.yaxis.set_major_locator(
            MaxNLocator(nbins=num_ticks, prune='both'))

        axhisty.xaxis.set_major_locator(
            MaxNLocator(nbins=num_ticks, prune='both'))

        if not self.xfirst_ax and hist_share:
            self.xfirst_ax = axhistx

        if not self.yfirst_ax and hist_share:
            self.yfirst_ax = axhisty

        # Keep track of which axes are which, because looking into fig.axes
        # list will get awkward....
        self.top_hists.append(axhistx)
        self.right_hists.append(axhisty)

        # Scatter will deal with NaN, but hist will not.  So clean the data
        # here.
        hx = _clean(x)
        hy = _clean(y)

        self.hxs.append(hx)
        self.hys.append(hy)

        # Only plot hists if there's valid data
        if len(hx) > 0:
            axhistx.hist(hx, **xhk)
        if len(hy) > 0:
            axhisty.hist(hy, **yhk)

        # Turn off unnecessary labels -- for these, use the scatter's axes
        # labels
        for txt in axhisty.get_yticklabels() + axhistx.get_xticklabels():
            txt.set_visible(False)

        for txt in axhisty.get_xticklabels():
            txt.set_rotation(-90)

    def add_legends(self, xhists=True, yhists=False, scatter=True, **kwargs):
        """
        Add legends to axes.
        """
        axs = []
        if xhists:
            axs.extend(self.hxs)
        if yhists:
            axs.extend(self.hys)
        if scatter:
            axs.extend(self.ax)

        for ax in axs:
            ax.legend(**kwargs)



import copy
from textwrap import dedent
import numpy as np
import pandas
from matplotlib import pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
import matplotlib.patches as patches
import matplotlib
import plotutils
from matplotlib.transforms import blended_transform_factory
from matplotlib.collections import EventCollection
import gffutils
import pybedtools
from pybedtools import featurefuncs


class ResultsTable(object):
    def __init__(self, data, db=None, import_kwargs=None):
        """
        Wrapper around a pandas.DataFrame that adds additional functionality.

        The underlying pandas.DataFrame is always available with the `data`
        attribute.

        Any attributes not explicitly in ResultsTable will be looked for in the
        underlying pandas.DataFrame.

        Parameters
        ----------
        data : string or pandas.DataFrame
            If string, assumes it's a filename and calls
            pandas.read_table(data, **import_kwargs).

        db : string or gffutils.FeatureDB
            Optional database that can be used to generate features

        import_kwargs : dict
            These arguments will be passed to pandas.read_table() if `data` is
            a filename.
        """
        if isinstance(data, basestring):
            import_kwargs = import_kwargs or {}
            data = pandas.read_table(data, **import_kwargs)
        if not isinstance(data, pandas.DataFrame):
            raise ValueError("`data` is not a pandas.DataFrame")
        self.data = data

        self._kwargs = dict(db=db, import_kwargs=import_kwargs)
        self.attach_db(db)
        self._cached_features = None

    def __getattr__(self, attr):
        if attr in self.__dict__:
            return getattr(self, attr)
        return getattr(self.data, attr)

    def __getitem__(self, attr):
        if isinstance(attr, basestring):
            return self.data.__getitem__(attr)
        else:
            return self.__class__(self.data.__getitem__(attr), **self._kwargs)

    def update(self, dataframe):
        """
        Updates the current data with a new dataframe.

        This extra step is required to get around the fancy pandas.DataFrame
        indexing (like .ix, .iloc, etc).
        """
        return self.__class__(dataframe, **self._kwargs)

    def copy(self):
        data = self.data.copy(deep=True)
        return self.__class__(data, db=self.db, import_kwargs=self._kwargs)

    def __repr__(self):
        s = []
        s.append("<%s instance, wrapping the following:"
                 % self.__class__.__name__)
        s.append('')
        s.extend('\t' + i for i in repr(self.data).splitlines(False))
        s.append('>')
        return '\n'.join(s)

    def attach_db(self, db):
        """
        Attach a gffutils.FeatureDB for access to features.

        Useful if you want to attach a db after this instance has already been
        created.

        Parameters
        ----------
        db : gffutils.FeatureDB
        """
        if db is not None:
            if isinstance(db, basestring):
                db = gffutils.FeatureDB(db)
            if not isinstance(db, gffutils.FeatureDB):
                raise ValueError(
                    "`db` must be a filename or a gffutils.FeatureDB")
        self._kwargs['db'] = db
        self.db = db

    def features(self, ignore_unknown=False):
        """
        Generator of features.

        If a gffutils.FeatureDB is attached, returns a pybedtools.Interval for
        every feature in the dataframe's index.

        Parameters
        ----------
        ignore_unknown : bool
            If True, silently ignores features that are not found in the db.
        """
        if not self.db:
            raise ValueError("Please attach a gffutils.FeatureDB")
        for i in self.data.index:
            try:
                yield gffutils.helpers.asinterval(self.db[i])
            except gffutils.FeatureNotFoundError:
                if ignore_unknown:
                    continue
                else:
                    raise gffutils.FeatureNotFoundError('%s not found' % i)

    def reindex_to(self, x, attribute="Name"):
        """
        Returns a copy that only has rows corresponding to feature names in x.

        Parameters
        ----------
        x : str or pybedtools.BedTool
            BED, GFF, GTF, or VCF where the "Name" field (that is, the value
            returned by feature['Name']) or any arbitrary attribute

        attribute : str
            Attribute containing the name of the feature to use as the index.
        """
        names = [i[attribute] for i in x]
        new = self.copy()
        new.data = new.data.reindex(names)
        return new

    def five_prime(self, upstream=1, downstream=0):
        """
        Creates a BED/GFF file of the 5' end of each feature represented in the
        table and returns the resulting pybedtools.BedTool object. Needs an
        attached database.

        Parameters
        ----------
        upstream, downstream : int
            Number of basepairs up and downstream to include
        """
        return pybedtools.BedTool(self.features())\
            .each(featurefuncs.five_prime, upstream, downstream)\
            .saveas()

    def three_prime(self, upstream=0, downstream=1):
        """
        Creates a BED/GFF file of the 3' end of each feature represented in the
        table and returns the resulting pybedtools.BedTool object. Needs an
        attached database.

        Parameters
        ----------
        upstream, downstream : int
            Number of basepairs up and downstream to include
        """
        return pybedtools.BedTool(self.features())\
            .each(featurefuncs.three_prime, upstream, downstream)\
            .saveas()

    TSS = five_prime
    TTS = three_prime

    def align_with(self, other):
        """
        Align the dataframe's index with another.
        """
        return self.__class__(self.data.reindex_like(other), **self._kwargs)

    def genes_in_common(self, other):
        """
        Convenience method for getting the genes found in both dataframes.
        """
        return self.index & other.index

    def __and__(self, other):
        return self.index & other.index

    def __or__(self, other):
        return self.index | other.index

    def __sub__(self, other):
        return self.index - other.index

    def scatter(self, x, y, xfunc=None, yfunc=None, xscale=None, yscale=None,
                xlab=None, ylab=None, genes_to_highlight=None,
                label_genes=False, marginal_histograms=False,
                general_kwargs=dict(color="k", alpha=0.2, linewidths=0),
                general_hist_kwargs=None,
                offset_kwargs={}, label_kwargs=None, ax=None,
                one_to_one=None, callback=None, xlab_prefix=None,
                ylab_prefix=None, sizefunc=None, hist_size=0.3, hist_pad=0.0,
                nan_offset=0.015, pos_offset=0.99, linelength=0.01,
                neg_offset=0.005):
        """
        Do-it-all method for making annotated scatterplots.

        Parameters
        ----------

        x, y : array-like
            Variables to plot.  Must be names in self.data's DataFrame.  For
            example, "baseMeanA" and "baseMeanB"

        xfunc, yfunc : callable
            Functions to apply to `xvar` and `yvar` respectively. Default is
            log2; set to None to have no transformation.

        xlab, ylab : string
            Labels for x and y axes; default is to use function names for
            `xfunc` and `yfunc` and variable names `xvar` and `yvar`, e.g.,
            "log2(baseMeanA)"

        ax : None or Axes object
            If `ax=None`, then makes a new fig and returns the Axes object,
            otherwise, plots onto `ax`

        general_kwargs : dict
            Kwargs for matplotlib.scatter; specifies how all points look

        genes_to_highlight : list of (index, dict) tuples
            Provides lots of control to colors.  It is a list of (`ind`,
            `kwargs`) tuples, where each `ind` specifies genes to plot with
            `kwargs`.  Each dictionary updates a copy of `general_kwargs`. If
            `genes_to_highlight` has a "name" kwarg, this must be a list that't
            the same length as `ind`.  It will be used to label the genes in
            `ind` using `label_kwargs`.

        callback : callable
            Function to call upon clicking a point. Default is to print the
            gene name, but an example of another useful callback would be
            a mini-browser connected to a genomic_signal object from which the
            expression data were calculated.

        one_to_one : None or dict
            If not None, a dictionary of matplotlib.plot kwargs that will be
            used to plot a 1:1 line.

        label_kwargs : dict
            Kwargs for labeled genes (e.g., dict=(style='italic')).  Will only
            be used if an entry in `genes_to_highlight` has a `name` key.

        offset_kwargs : dict
            Kwargs to be passed to matplotlib.transforms.offset_copy, used for
            adjusting the positioning of gene labels in relation to the actual
            point.

        xlab_prefix, ylab_prefix : str
            Optional label prefix that will be added to the beginning of `xlab`
            and/or `ylab`.

        hist_size : float
            Size of marginal histograms

        hist_pad : float
            Spacing between marginal histograms

        nan_offset, pos_offset, neg_offset : float
            Offset, in units of "fraction of axes" for the NaN, +inf, and -inf
            "rug plots"

        linelength : float
            Line length for the rug plots

        """
        _x = self.data[x]
        _y = self.data[y]

        # Construct defaults---------------------------------------------------
        def identity(x):
            return x.copy()

        # Axis label setup
        if xlab_prefix is None:
            xlab_prefix = ""

        if ylab_prefix is None:
            ylab_prefix = ""

        if xlab is None:
            xlab = x

            if xfunc is not None:
                xlab = xlab_prefix + "%s(%s)" % (xfunc.__name__, str(x))
            else:
                xlab = xlab_prefix + "%s" % (str(x))

        if ylab is None:
            ylab = y
            if yfunc is not None:
                ylab = ylab_prefix + "%s(%s)" % (yfunc.__name__, str(y))
            else:
                ylab = ylab_prefix + "%s" % (str(y))

        if xfunc is None:
            xfunc = identity

        if yfunc is None:
            yfunc = identity

        if general_kwargs is None:
            general_kwargs = {}

        if general_hist_kwargs is None:
            general_hist_kwargs = {}

        if genes_to_highlight is None:
            genes_to_highlight = []

        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(111)

        if label_kwargs is None:
            label_kwargs = dict(horizontalalignment='right',
                                verticalalignment='center', style='italic',
                                bbox=dict(facecolor='w', edgecolor='None',
                                          alpha=0.5))

        # Clean data ---------------------------------------------------------
        xi = xfunc(_x)
        yi = yfunc(_y)

        # handle inf, -inf, and NaN
        x_is_pos_inf = np.isinf(xi) & (xi > 0)
        x_is_neg_inf = np.isinf(xi) & (xi < 0)
        x_is_nan = np.isnan(xi)
        y_is_pos_inf = np.isinf(yi) & (yi > 0)
        y_is_neg_inf = np.isinf(yi) & (yi < 0)
        y_is_nan = np.isnan(yi)

        # Indexes for valid values
        x_valid = ~(x_is_pos_inf | x_is_neg_inf | x_is_nan)
        y_valid = ~(y_is_pos_inf | y_is_neg_inf | y_is_nan)

        gmin = max(xi[x_valid].min(), yi[y_valid].min())
        gmax = min(xi[x_valid].max(), yi[y_valid].max())


        # Convert any integer indexes into boolean, and create a new list of
        # genes to highlight.  This handles optional hist kwargs.
        allind = np.zeros_like(xi) == 0
        _genes_to_highlight = []
        for block in genes_to_highlight:
            ind = block[0]
            if ind.dtype != 'bool':
                new_ind = (np.zeros_like(xi) == 0)
                new_ind[ind] = True
                _genes_to_highlight.append(
                    tuple([new_ind] + list(block[1:]))
                )
            else:
                _genes_to_highlight.append(
                    tuple([ind] + list(block[1:]))
                )

        # Remove any genes that will be plotted by genes_to_highlight.  This
        # avoids double-plotting.
        for block in _genes_to_highlight:
            ind = block[0]
            allind[ind] = False

        # Copy over the color and alpha if they're not specified
        general_hist_kwargs = plotutils._updatecopy(
            orig=general_hist_kwargs, update_with=general_kwargs,
            keys=['color', 'alpha'])

        # Put the non-highlighted genes at the beginning of _genes_to_highlight
        # so we can just iterate over that.
        _genes_to_highlight.insert(
            0,
            (allind, general_kwargs, general_hist_kwargs)
        )

        # Set up the object that will handle the marginal histograms
        self.marginal = plotutils.MarginalHistScatter(
            ax, hist_size=hist_size, pad=hist_pad)


        # Set up kwargs for x and y rug plots
        rug_x_kwargs = dict(
            linelength=linelength,
            transform=blended_transform_factory(ax.transData, ax.transAxes),
        )
        rug_y_kwargs = dict(
            linelength=linelength,
            transform=blended_transform_factory(ax.transAxes, ax.transData),
            orientation='vertical',
        )

        # EventCollection objects need a color as a 3-tuple, so set up
        # a converter here.
        color_converter = matplotlib.colors.ColorConverter().to_rgb

        # Plot the one-to-one line, if kwargs were specified
        if one_to_one:
            ax.plot([gmin, gmax],
                    [gmin, gmax],
                    **one_to_one)

        # Plot any specially-highlighted genes, and label if specified
        for block in _genes_to_highlight:
            ind = block[0]
            kwargs = block[1]

            if len(block) == 3:
                hist_kwargs = block[2]
            else:
                hist_kwargs = {}

            names = kwargs.pop('names', None)
            _marginal_histograms = (
                kwargs.pop('marginal_histograms', False) or
                marginal_histograms)

            updated_kwargs = plotutils._updatecopy(
                orig=kwargs, update_with=general_kwargs)

            updated_hist_kwargs = plotutils._updatecopy(
                orig=hist_kwargs, update_with=general_hist_kwargs)
            updated_hist_kwargs = plotutils._updatecopy(
                orig=updated_hist_kwargs, update_with=kwargs,
                keys=['color', 'alpha'], override=True)

            xhist_kwargs = updated_kwargs.pop('xhist_kwargs', None)
            yhist_kwargs = updated_kwargs.pop('yhist_kwargs', None)

            self.marginal.append(
                xi[ind & x_valid & y_valid],
                yi[ind & x_valid & y_valid],
                scatter_kwargs=dict(picker=5, **updated_kwargs),
                hist_kwargs=updated_hist_kwargs,
                xhist_kwargs=xhist_kwargs,
                yhist_kwargs=yhist_kwargs,
                marginal_histograms=_marginal_histograms)
            coll = self.marginal.scatter_ax.collections[-1]
            coll.df = self.data
            coll.ind = ind

            color = color_converter(updated_kwargs['color'])
            rug_x_kwargs['color'] = color
            rug_y_kwargs['color'] = color

            # Note: if both x and y are not valid, then they will not be on the
            # plot.
            items = [
                # top rug, y is +inf and x is valid
                (xi, ind & x_valid & y_is_pos_inf, pos_offset, rug_x_kwargs),

                # one of the bottom rugs, where y is NaN
                (xi, ind & x_valid & y_is_nan, nan_offset, rug_x_kwargs),

                # bottom rug, y is -inf
                (xi, ind & x_valid & y_is_neg_inf, neg_offset, rug_x_kwargs),

                # right rug, x is +inf
                (yi, ind & y_valid & x_is_pos_inf, pos_offset, rug_y_kwargs),

                # one of the left rugs; x is NaN
                (yi, ind & y_valid & x_is_nan, nan_offset, rug_y_kwargs),

                # left rug, x is -inf
                (yi, ind & y_valid & x_is_neg_inf, neg_offset, rug_y_kwargs),
            ]
            for values, index, offset, kwargs in items:
                coll = EventCollection(
                    values[index], lineoffset=offset, **kwargs)
                coll.df = self.data
                coll.ind = index
                ax.add_collection(coll)

            if names:
                transOffset = matplotlib.transforms.offset_copy(
                    ax.transData, fig=ax.figure, **offset_kwargs)

                for xii, yii, name in zip(xi[ind], yi[ind], names):
                    ax.text(xii,
                            yii,
                            name,
                            transform=transOffset,
                            **label_kwargs)

        # register callback
        if callback is not None:
            def wrapped_callback(event):
                return callback(self._id_callback(event))

        else:
            def wrapped_callback(event):
                return self._default_callback(self._id_callback(event))

        ax.figure.canvas.mpl_connect('pick_event', wrapped_callback)

        ax.set_xlabel(xlab)
        ax.set_ylabel(ylab)
        ax.axis('tight')

        #ax.axis((xmin - xpad, xmax + xpad, ymin - ypad, ymax + ypad))
        return ax


    def radviz(self, column_names, transforms=dict(), **kwargs):
        """
        Radviz plot.

        Useful for exploratory visualization, a radviz plot can show
        multivariate data in 2D.  Conceptually, the variables (here, specified
        in `column_names`) are distributed evenly around the unit circle.  Then
        each point (here, each row in the dataframe) is attached to each
        variable by a spring, where the stiffness of the spring is proportional
        to the value of corresponding variable.  The final position of a point
        represents the equilibrium position with all springs pulling on it.

        In practice, each variable is normalized to 0-1 (by subtracting the
        mean and dividing by the range).

        This is a very exploratory plot.  The order of `column_names` will
        affect the results, so it's best to try a couple different orderings.
        For other caveats, see [1].

        Additional kwargs are passed to self.scatter, so subsetting, callbacks,
        and other configuration can be performed using options for that method
        (e.g., `genes_to_highlight` is particularly useful).

        Parameters
        ----------
        column_names : list
            Which columns of the dataframe to consider.  The columns provided
            should only include numeric data, and they should not contain any
            NaN, inf, or -inf values.

        transforms : dict
            Dictionary mapping column names to transformations that will be
            applied just for the radviz plot.  For example, np.log1p is
            a useful function. If a column name is not in this dictionary, it
            will be used as-is.

        ax : matplotlib.Axes
            If not None, then plot the radviz on this axes.  If None, then
            a new figure will be created.

        kwargs : dict
            Additional arguments are passed to self.scatter.  Note that not all
            possible kwargs for self.scatter are necessarily useful for
            a radviz plot (for example, margninal histograms would not be
            meaningful).

        Notes
        -----
        This method adds two new variables to self.data: "radviz_x" and
        "radviz_y".  It then calls the self.scatter method, using these new
        variables.

        The data transformation was adapted from the
        pandas.tools.plotting.radviz function.

        References
        ----------
        [1]  Hoffman,P.E. et al. (1997) DNA visual and analytic data mining. In
             the Proceedings of the IEEE Visualization. Phoenix, AZ, pp. 437-441.
        [2] http://www.agocg.ac.uk/reports/visual/casestud/brunsdon/radviz.htm
        [3] http://pandas.pydata.org/pandas-docs/stable/visualization.html#radviz
        """
        # make a copy of data
        x = self.data[column_names].copy()

        for k, v in transforms.items():
            x[k] = v(x[k])

        def normalize(series):
            mn = min(series)
            mx = max(series)
            return (series - mn) / (mx - mn)

        df = x.apply(normalize)

        to_plot = []

        n = len(column_names)

        s = np.array([(np.cos(t), np.sin(t))
                      for t in [2.0 * np.pi * (i / float(n))
                                for i in range(n)]])

        for i in range(len(x)):
            row = df.irow(i).values
            row_ = np.repeat(np.expand_dims(row, axis=1), 2, axis=1)
            to_plot.append((s * row_).sum(axis=0) / row.sum())

        x_, y_ = zip(*to_plot)
        self.data['radviz_x'] = x_
        self.data['radviz_y'] = y_

        ax = self.scatter('radviz_x', 'radviz_y', **kwargs)

        # Thanks to 
        ax.add_patch(patches.Circle((0.0, 0.0), radius=1.0, facecolor='none'))
        for xy, name in zip(s, column_names):
            ax.add_patch(patches.Circle(xy, radius=0.025, facecolor='gray'))
            if xy[0] < 0.0 and xy[1] < 0.0:
                ax.text(xy[0] - 0.025, xy[1] - 0.025, name,
                        ha='right', va='top', size='small')
            elif xy[0] < 0.0 and xy[1] >= 0.0:
                ax.text(xy[0] - 0.025, xy[1] + 0.025, name,
                        ha='right', va='bottom', size='small')
            elif xy[0] >= 0.0 and xy[1] < 0.0:
                ax.text(xy[0] + 0.025, xy[1] - 0.025, name,
                        ha='left', va='top', size='small')
            elif xy[0] >= 0.0 and xy[1] >= 0.0:
                ax.text(xy[0] + 0.025, xy[1] + 0.025, name,
                        ha='left', va='bottom', size='small')

        ax.axis('equal')
        return ax



    def _id_callback(self, event):
        for i in event.ind:
            return event.artist.df.ix[event.artist.ind].index[i]

    def _default_callback(self, i):
        print self.data.ix[i]

    def strip_unknown_features(self):
        """
        Remove features not found in the `gffutils.FeatureDB`.  This will
        typically include 'ambiguous', 'no_feature', etc, but can also be
        useful if the database was created from a different one than was used
        to create the table.
        """
        if not self.db:
            return self
        ind = []
        for i, gene_id in enumerate(self.data.index):
            try:
                self.db[gene_id]
                ind.append(i)
            except gffutils.FeatureNotFoundError:
                pass
        ind = np.array(ind)
        return self.__class__(self.data.ix[ind], **self._kwargs)

    def genes_with_peak(self, peaks, transform_func=None, split=False,
                        intersect_kwargs=None, *args, **kwargs):
        """
        Returns a boolean index of genes that have a peak nearby.

        Parameters
        ----------
        peaks : string or pybedtools.BedTool
            If string, then assume it's a filename to a BED/GFF/GTF file of
            intervals; otherwise use the pybedtools.BedTool object directly.

        transform_func : callable
            This function will be applied to each gene object returned by
            self.features().  Additional args and kwargs are passed to
            `transform_func`. For example, if you're looking for peaks within
            1kb upstream of TSSs, then pybedtools.featurefuncs.TSS would be
            a useful `transform_func`, and you could supply additional kwargs
            of `upstream=1000` and `downstream=0`.

            This function can return iterables of features, too. For example,
            you might want to look for peaks falling within the exons of
            a gene.  In this case, `transform_func` should return an iterable
            of pybedtools.Interval objects.  The only requirement is that the
            `name` field of any feature matches the index of the dataframe.

        intersect_kwargs : dict
            kwargs passed to pybedtools.BedTool.intersect.
        """
        def _transform_func(x):
            """
            In order to support transform funcs that return a single feature or
            an iterable of features, we need to wrap it
            """
            result = transform_func(x)
            if isinstance(result, pybedtools.Interval):
                result = [result]
            for i in result:
                if i:
                    yield result


        intersect_kwargs = intersect_kwargs or {}
        if not self._cached_features:
            self._cached_features = pybedtools\
                .BedTool(self.features())\
                .saveas()

        if transform_func:
            if split:
                features = self._cached_features\
                    .split(_transform_func, *args, **kwargs)
            else:
                features = self._cached_features\
                    .each(transform_func, *args, **kwargs)

        else:
            features = self._cached_features

        hits = list(set([i.name for i in features.intersect(peaks, **intersect_kwargs)]))
        return self.data.index.isin(hits)


class DESeqResults(ResultsTable):
    def __init__(self, data, db=None, header_check=True,
                 remove_deseq_extra=True, **kwargs):

        to_remove = [
            'not_aligned',
            'no_feature',
            'ambiguous',
            'alignment_not_unique',
            'too_low_aQual']

        import_kwargs = kwargs.pop('import_kwargs', {})
        if header_check and isinstance(data, basestring):
            comment_char = import_kwargs.get('comment', '#')
            for i, line in enumerate(open(data)):
                if line[0] != comment_char:
                    break
            import_kwargs['skiprows'] = i
        import_kwargs['na_values'] = ['nan']

        import_kwargs['index_col'] = import_kwargs.pop('index_col', 0)
        super(DESeqResults, self).__init__(
            data=data, db=db, import_kwargs=import_kwargs, **kwargs)
        self.data.log2FoldChange = self.data.log2FoldChange.astype(float)
        self.data.foldChange = self.data.foldChange.astype(float)
        if remove_deseq_extra:
            self.data = self.data[~self.data.index.isin(to_remove)]

    def changed(self, thresh=0.05, idx=True, col='padj'):
        """
        Changed features.

        {threshdoc}
        """
        ind = self.data[col] <= thresh
        if idx:
            return ind
        return self[ind]

    def unchanged(self, thresh=0.05, idx=True, col='padj'):
        """
        Changed features.

        {threshdoc}
        """
        ind = (self.data[col] > thresh) | np.isnan(self.data[col])
        if idx:
            return ind
        return self[ind]

    def enriched(self, thresh=0.05, idx=True, col='padj'):
        """
        Enriched features.

        {threshdoc}
        """
        return self.upregulated(thresh=thresh, idx=idx)

    def upregulated(self, thresh=0.05, idx=True, col='padj'):
        """
        Upregulated features.

        {threshdoc}
        """
        ind = (self.data[col] <= thresh) & (self.data['log2FoldChange'] > 0)
        if idx:
            return ind
        return self[ind]

    def downregulated(self, thresh=0.05, idx=True, col='padj'):
        """
        Downregulated features.

        {threshdoc}
        """
        ind = (self.data[col] <= thresh) & (self.data['log2FoldChange'] < 0)
        if idx:
            return ind
        return self[ind]

    def disenriched(self, thresh=0.05, idx=True, col='padj'):
        """
        Disenriched features.

        {threshdoc}
        """
        return self.downregulated(thresh=thresh, idx=idx)

    def ma_plot(self, thresh, col='padj', up_kwargs=None, dn_kwargs=None,
                zero_line=None, **kwargs):
        """
        MA plot

        Plots the average read count across treatments (x-axis) vs the log2
        fold change (y-axis).

        Additional kwargs are passed to self.scatter (useful ones might include
        `genes_to_highlight`)

        Parameters
        ----------
        thresh : float
            Features with values <= `thresh` will be highlighted in the plot.

        col : str
            Column name that `thresh` will be applied to

        up_kwargs, dn_kwargs : None or dict
            Kwargs passed to matplotlib's scatter(), used for styling up/down
            regulated features (defined by `thresh` and `col`)

        zero_line : None or dict
            Kwargs passed to matplotlib.axhline(0).

        """
        genes_to_highlight = kwargs.pop('genes_to_highlight', [])
        genes_to_highlight.append(
            (self.upregulated(thresh, col=col),
             up_kwargs or dict(color='r')))
        genes_to_highlight.append(
            (self.downregulated(thresh, col=col),
             dn_kwargs or dict(color='b')))
        if zero_line is None:
            zero_line = {}
        x = 'baseMean'
        y = 'log2FoldChange'

        if 'xfunc' not in kwargs:
            xfunc = np.log1p
        ax = self.scatter(
            x=x,
            y=y,
            xfunc=xfunc,
            genes_to_highlight=genes_to_highlight,
            **kwargs)
        if zero_line:
            ax.axhline(0, **zero_line)
        return ax

    def colormapped_bedfile(self, genome, cmap=None):
        """
        Create a BED file with padj encoded as color

        Features will be colored according to adjusted pval (phred
        transformed).  Downregulated features have the sign flipped.

        Parameters
        ----------
        cmap : matplotlib colormap
            Default is matplotlib.cm.RdBu_r

        Notes
        -----
        Requires a FeatureDB to be attached.
        """
        if self.db is None:
            raise ValueError("FeatureDB required")
        db = gffutils.FeatureDB(self.db)

        def scored_feature_generator(d):
            for i in range(len(d)):
                try:
                    feature = db[d.ix[i]]
                except gffutils.FeatureNotFoundError:
                    raise gffutils.FeatureNotFoundError(d.ix[i])
                score = -10 * np.log10(d.padj[i])
                lfc = d.log2FoldChange[i]
                if np.isnan(lfc):
                    score = 0
                if lfc < 0:
                    score *= -1
                feature.score = str(score)
                feature = extend_fields(
                    gff2bed(gffutils.helpers.asinterval(feature)), 9)
                fields = feature.fields[:]
                fields[6] = fields[1]
                fields[7] = fields[2]
                fields.append(str(d.padj[i]))
                fields.append(str(d.pval[i]))
                fields.append('%.3f' % d.log2FoldChange[i])
                fields.append('%.3f' % d.baseMeanB[i])
                fields.append('%.3f' % d.baseMeanB[i])
                yield pybedtools.create_interval_from_list(fields)

        x = pybedtools.BedTool(scored_feature_generator(self)).saveas()
        norm = x.colormap_normalize()
        if cmap is None:
            cmap = cm.RdBu_r
        cmap = colormap_adjust.cmap_center_point_adjust(
            cmap, [norm.vmin, norm.vmax], 0)

        def score_zeroer(f):
            f.score = '0'
            return f
        return x.each(add_color, cmap=cmap, norm=norm)\
                .sort()\
                .each(score_zeroer)\
                .truncate_to_chrom(genome)\
                .saveas()

    def autosql_file(self):
        """
        Generate the autosql for DESeq results (to create bigBed)

        Returns a temp filename containing the autosql defining the extra
        fields.

        This for creating bigBed files from BED files created by
        colormapped_bed.  When a user clicks on a feature, the DESeq results
        will be reported.
        """
        fn = pybedtools.BedTool._tmp()

        AUTOSQL = dedent(
            """
            table example
            "output from DESeq"
            (
            string  chrom;  "chromosome"
            uint chromStart; "start coord"
            uint chromEnd; "stop coord"
            string name; "name of feature"
            uint score; "always zero"
            char[1] strand; "+ or - for strand"
            uint    thickStart; "Coding region start"
            uint    thickEnd;  "Coding region end"
            uint reserved; "color according to score"
            string padj; "DESeq adjusted p value"
            string pval; "DESeq raw p value"
            string logfoldchange; "DESeq log2 fold change"
            string basemeana; "DESeq baseMeanA"
            string basemeanb; "DESeq baseMeanB"
        )
        """)

        fout = open(fn, 'w')
        fout.write(AUTOSQL)
        fout.close()
        return fn

    threshdoc = """
    Parameters
    ----------
    thresh : float
        Only features with <= `thresh` will be returned

    idx : bool
        If True, a boolean index will be returned.  If False, a new object will
        be returned that has been subsetted.

    col : str
        Name of the column to apply `thresh` to
    """
    enriched.__doc__ = enriched.__doc__.format(threshdoc=threshdoc)
    disenriched.__doc__ = disenriched.__doc__.format(threshdoc=threshdoc)
    upregulated.__doc__ = upregulated.__doc__.format(threshdoc=threshdoc)
    downregulated.__doc__ = downregulated.__doc__.format(threshdoc=threshdoc)


class LazyDict(object):
    def __init__(self, fn_dict, dbfn, index_file, extra=None, cls=DESeqResults,
                 modifier=None):
        """
        Dictionary-like object that lazily-loads ResultsTable objects.

        Parameters
        ----------
        fn_dict : dict
            Keys of `fn_dict` will be the keys of this LazyDict object.  Values
            should be filenames which will be loaded into ResultsTable object
            upon access for the first time.

        index_file : str
            Path to a file that contains one ID per line.  This file is used to
            ensure all ResultsTable objects are aligned to the same index.

        dbfn : str
            Filename to a gffutils database.  This enables gene info to be
            attached to the dataframe.

        extra : pandas.dataframe
            This dataframe hat will be merged into the data in each file.  This
            is useful for attaching things like gene lengths, alt names, etc.
            In order for it to work, this dataframe must be indexed the same
            way the ResultsTable files are indexed.

        cls : ResultsTable class or subclass
            Each filename in `fn_dict` will be converted using this class.

        modifier : callable
            Upon first access, each newly-constructed ResultsTable will first
            have the `extra` data attached, and then will be provided as this
            callable's only argument.  The callable can make any modifications
            to the ResultsTable, and return a new version that will be used in
            the future when the same key is accessed.  For example, exonic bp
            data can be provided as part of the `extra` dataframe, and then the
            `modifier` can be a function that adds an RPKM column.

        Notes
        -----
        When a key is provided for the first time, the workflow is
        ResultsTable(fn, **kwargs) -> attach `extra` -> send to `modifier` ->
        return extended and modified ResultsTable.  Subsequent access of the
        same key will immediately return the extended-and-modified
        ResultsTable.

        """
        self.fn_dict = fn_dict
        self._dict = {}
        self.dbfn = dbfn
        self.index = [i.strip() for i in open(index_file)]
        self.extra = extra.ix[self.index]
        self.modifier = modifier
        self._cls = cls

    def __getitem__(self, key):
        if key not in self._dict:
            fn = self.fn_dict[key]
            obj = self._cls(fn, db=self.dbfn)
            obj.data = obj.data.ix[self.index]
            if self.extra is not None:
                obj.data = pandas.merge(obj.data, self.extra, left_index=True,
                                        right_index=True)
            if self.modifier:
                obj = self.modifier(obj)

            self._dict[key] = obj

        return self._dict[key]

    def __repr__(self):
        s = "<%s> with possible keys\n:%s\n" \
            % (self.__class__.__name__, self.fn_dict.keys())
        s += "and existing keys:\n"
        s += repr(self._dict)
        return s

    def keys(self):
        return self.fn_dict.keys()

    def values(self):
        return [self._dict[key] for key in self.keys()]

    def items(self):
        return list((key, self._dict[key]) for key in self.keys())

if __name__ == "__main__":
    import metaseq
    from matplotlib import pyplot as plt

    db = metaseq.example_filename('dmel-all-r5.33-cleaned.gff.db')
    import_kwargs = dict(comment='#')
    d = DESeqResults(
        metaseq.example_filename('rrp6-s2-polyA.final.summary'),
        db=db,
        import_kwargs=import_kwargs,
    )

    e = DESeqResults(
        metaseq.example_filename('rrp40-s2-polyA.final.summary'),
        db=db,
        import_kwargs=import_kwargs,
    )

    d = d.align_with(e)

    en = e.enriched()

    from metaseq.minibrowser import BaseMiniBrowser, GeneModelMiniBrowser
    minibrowser = GeneModelMiniBrowser([], e.db)

    def callback(i):
        feature = gffutils.helpers.asinterval(e.db[i])
        print e.ix[i]
        if feature.chrom == 'chr2L':
            minibrowser.plot(feature)

    e.scatter(
        'baseMeanA',
        'baseMeanB',
        genes_to_highlight=[
            (en, dict(color='r', alpha=1.0)),
        ],
        xfunc=np.log1p,
        yfunc=np.log1p,
        one_to_one=dict(color='b', linestyle=':'),
        callback=callback,
    )

    e.ma_plot(
        0.05,
        zero_line=dict(color='y', linestyle=':'),
        genes_to_highlight=[
            (e.data.index.isin(['FBgn0041721']), dict(color='b', s=50))
        ],
    )

    plt.show()



from metaseq import results_table
import metaseq
import numpy as np

fn = metaseq.example_filename('ex.deseq')
d = results_table.ResultsTable(fn)


def test_dataframe_access():

    # different ways of accessing get the same data in memory
    assert d.id is d.data.id
    assert d['id'] is d.data.id

def test_dataframe_subsetting():
    assert all(d[:10].data == d.data[:10])
    assert all(d.update(d.data[:10]).data == d.data[:10])

def test_copy():
    e = d.copy()
    e.id = 'a'
    assert e.id[0] == 'a'
    assert d.id[0] != 'a'

def smoke_tests():
    #smoke test for repr
    print repr(d)

def test_db():

    # should work
    d.attach_db(None)

    d.attach_db(metaseq.example_filename('dmel-all-r5.33-cleaned.gff.db'))



import scipy
from scipy import signal
import numpy as np
import metaseq
import pybedtools
from matplotlib import pyplot as plt

x = metaseq.genomic_signal(
    'LM09.fastq.clipped.bowtie.bam.filtered.nodups', 'bam')

arr1 = x.local_coverage('chr2L:1-10000000', processes=8, fragment_size=1, read_strand='-')[1]
arr2 = x.local_coverage('chr2L:1-10000000', processes=8, fragment_size=1, read_strand='+')[1]

cc = signal.fftconvolve(arr1, arr2[::-1], mode='full')
center = len(cc)/2
normalized = cc / cc.max()#[center]
xi = np.arange(-center, center + 1)
print xi[np.argmax(normalized)]

plt.plot(xi, normalized)

b, a = signal.butter(3, 0.05)
filtered = signal.filtfilt(b, a, cc)

print xi[np.argmax(filtered)]

plt.plot(xi, filtered/filtered.max())

plt.axis(xmin=0, xmax=500)
plt.show()

# copied from scipy.signal.fftconvolve
"""
s1 = np.array(arr1.shape)
s2 = np.array(arr2.shape)
size = s1 + s2 - 1

# calculate closes power-of-2; fftn() will auto-pad to this size.
fsize = 2 ** np.ceil(np.log2(size))
fslice = tuple([slice(0, int(sz)) for sz in size])
res = np.fftn(arr1, fsize)
res *= np.fftn(arr2, fsize)
"""



import metaseq

x = metaseq.genomic_signal(
    metaseq.example_filename('sh_chr2L.bam'), 'bam')

#res = x.local_coverage(['chr2L:60000-70000', 'chr2L:80000-90000'], fragment_size=100, bins=[50, 10], processes=None)
res = x.local_coverage('chr2L:60000-170000', fragment_size=300, bins=100, processes=4)
#res = x.local_coverage(['chr2L:60000-70000', 'chr2L:80000-90000'], fragment_size=300, bins=[50, 1000], processes=None)

x, y = res
plot(x,y, '.-')
show()





"""
Module for spawning mini genome browsers using a plugin structure, making it
possible to build rather complex mini-browsers.  The goal is to point the
mini-browser to some data, and call its plot() method with a feature.  This
will spawn a new figure showing the data for that interval.

MiniBrowser classes are just a general way of mapping data-manipulation or
data-visualization methods to an Axes on which the data should be displayed.

To make a new subclass:

1. Create one or more methods that accept an Axes object and a pybedtools
   Interval object and return a feature.  The simplest do-nothing method would
   be::

        def my_panel(self, ax, feature)
            return feature

   A more useful method might be one that plots genomic signal over the
   region::

        def my_panel(self, ax, feature):
            # for simplicity, assume just use the first genomic_signal
            gs = self.genomic_signal_objs[0]
            x, y = gs.local_coverage(feature, bins=100)
            ax.plot(x, y, **kwargs)
            ax.axis('tight')
            return feature

2. Then, override the `panels()` method.  This method:

    * Creates Axes as needed; assumes that self.make_fig() has already been
      called so that self.fig is available.
    * Returns a list of (ax, method) tuples.  This list maps created Axes to
      methods that should operate on them (like `my_panel` method above).

    For example::

        def panels(self):
            ax = self.fig.add_subplot(111)
            return [(ax, self.my_panel)]


A figure is spawned by calling the `plot` method on a pybedtools genomic
interval, e.g.::

    s = SignalMiniBrowser(ip, control])
    s.plot(feature)

"""
from copy import copy
from matplotlib import pyplot as plt
from pybedtools.contrib.plotting import Track
import pybedtools
import gffutils
from gffutils.helpers import asinterval
from mpl_toolkits.axes_grid1 import make_axes_locatable
import matplotlib.gridspec as gridspec
from matplotlib.ticker import FormatStrFormatter
import _genomic_signal


class BaseMiniBrowser(object):
    """
    Base class for plotting a genomic region.

    This class is designed to be sub-classed, so it really just a shell with
    some example methods filled in.
    """
    _all_figures = []

    def __init__(self, genomic_signal_objs):
        self.genomic_signal_objs = genomic_signal_objs

    def close_all(self):
        """
        Close all figures spawned by this class.
        """
        for f in self._all_figures:
            plt.close(f)

    def plot(self, feature):
        """
        Spawns a new figure showing data for `feature`.

        :param feature: A `pybedtools.Interval` object

        Using the pybedtools.Interval `feature`, creates figure specified in
        :meth:`BaseMiniBrowser.make_fig` and plots data on panels according to
        `self.panels()`.
        """
        if isinstance(feature, gffutils.Feature):
            feature = asinterval(feature)
        self.make_fig()
        axes = []
        for ax, method in self.panels():
            feature = method(ax, feature)
            axes.append(ax)
        return axes

    def make_fig(self):
        """
        Figure constructor, called before `self.plot()`
        """
        self.fig = plt.figure(figsize=(8, 4))
        self._all_figures.append(self.fig)

    def panels(self):
        """
        Method to be overriden by subclasses.

        This method should create Axes objects on self.fig and register which
        methods operate on which axes.

        Each method's signature should be func(self, ax, feature), and it will
        take a pybedtools.Interval `feature` and plot it or something derived
        from it onto `ax`. It should return a pybedtools.Interval object,
        usually just the same one that was passed in.
        """
        ax = self.fig.add_subplot(111)
        return [(ax, self.example_panel)]

    def example_panel(self, ax, feature):
        """
        A example panel that just prints the text of the feature.
        """
        txt = '%s:%s-%s' % (feature.chrom, feature.start, feature.stop)
        ax.text(0.5, 0.5, txt, transform=ax.transAxes)
        return feature


class ChIPSeqMiniBrowser(BaseMiniBrowser):
    def __init__(self, ip, control, db=None, peaks=None,
                 local_coverage_kwargs=dict(stranded=False), ip_style=None,
                 control_style=None, peaks_style=None, max_bins=500):
        """
        Mini-browser to show IP and control signal, along with optional peaks
        and gene models.  See module-level documentation for details.

        Parameters
        ----------
        ip, control : genomic_signal object
            These can be any kind of genomic signal object.  If BamSignal, then
            the signal will be scaled by million mapped reads.

        db : filename or gffutils.FeatureDB
            Optional database of annotations.  If provided, gene models will be
            plotted on an additional axes.

        peaks : filename or pybedtools.BedTool object
            Optional file of called peaks.  If provided, will be plotted on an
            additional axes.

        local_coverage_kwargs : dict
            Kwargs to pass on to the local_coverage method of `ip` and
            `control`.  It is recommended that at least `stranded=False` is
            included, since the mini-browser always displays the plus strand.

        ip_style, control_style : dict
            Kwargs to pass to Axes.fill_between (e.g. colors, alpha, line styles)

        peaks_style : dict
            Kwargs to pass to PolyCollection (e.g., face_color, edgewidth)

        max_bins : int
            Maximum number of bins to use when getting genomic_signal.  This can
            be overridden by providing the `bins` kwarg in
            `local_coverage_kwargs`.

        Notes
        -----
        After creation, settings can be changed in between calls to the
        `plot()` method.

        For example, the self.settings dictionary, which
        contains arguments passed to the gffutils.contrib.plotting.Gene class,
        can be modified to specify which kinds of features should be plotted.
        The `gs` attribute contains the matplotlib.gridspec.GridSpec object,
        which can be used to modify the relative sizes of the subplots.


        """
        super(ChIPSeqMiniBrowser, self).__init__([ip, control])
        self.local_coverage_kwargs = local_coverage_kwargs or {}
        self.ip = ip
        self.control = control
        if isinstance(db, basestring):
            db = gffutils.FeatureDB(db)
        self.db = db
        self.ip_style = ip_style or {}
        self.peaks_style = peaks_style or {}
        self.control_style = control_style or {}

        if peaks is not None:
            peaks = pybedtools.BedTool(peaks)
        self.peaks = peaks
        self.max_bins = max_bins

        self.settings = {
            'transcripts': None,
            'cds': ['CDS'],
            'utrs': ['exon'],
            'color': '0.5',
            'featuretype': 'gene',
        }

    def panels(self):
        self.fig.set_facecolor('w')
        method_dispatch = {
            'ip': self.ip_panel,
            'control': self.control_panel,
            'peaks': self.peaks_panel,
            'genes': self.genes_panel}
        if self.db and self.peaks:
            gs = gridspec.GridSpec(4, 1, height_ratios=[1, 1, .15, .5])
            ip_ax = plt.subplot(gs[0])
            control_ax = plt.subplot(gs[1], sharex=ip_ax, sharey=ip_ax)
            peaks_ax = plt.subplot(gs[2], sharex=ip_ax)
            gene_ax = plt.subplot(gs[3], sharex=ip_ax)
            axes = {'ip': ip_ax, 
                    'control': control_ax,
                    'peaks': peaks_ax,
                    'genes': gene_ax,
                    }


        elif self.db and self.peaks is None:
            gs = gridspec.GridSpec(3, 1, height_ratios=[1, 1, .5])
            ip_ax = plt.subplot(gs[0])
            control_ax = plt.subplot(gs[1], sharex=ip_ax, sharey=ip_ax)
            gene_ax = plt.subplot(gs[2], sharex=ip_ax)
            axes = {'ip': ip_ax, 
                    'control': control_ax,
                    'genes': gene_ax,
                    }


        elif self.db is None and self.peaks:
            gs = gridspec.GridSpec(3, 1, height_ratios=[1, 1, .3])
            ip_ax = plt.subplot(gs[0])
            control_ax = plt.subplot(gs[1], sharex=ip_ax, sharey=ip_ax)
            peaks_ax = plt.subplot(gs[2], sharex=ip_ax)
            axes = {'ip': ip_ax, 
                    'control': control_ax,
                    'peaks': peaks_ax,
                    }

        elif self.db is None and self.peaks is None:
            gs = gridspec.GridSpec(2, 1, height_ratios=[1, 1])
            ip_ax = plt.subplot(gs[0])
            control_ax = plt.subplot(gs[1], sharex=ip_ax, sharey=ip_ax)
            axes = {'ip': ip_ax, 
                    'control': control_ax,
                    }

        self.axes = axes
        self.gridspec = gs

        # Order shouldn't matter, but just in case...
        mapping = []
        for kind in ['ip', 'control', 'peaks', 'genes']:
            if kind in axes:
                mapping.append((axes[kind], method_dispatch[kind]))

        self._first_ax = mapping[0][0]
        self._last_ax = mapping[-1][0]

        return mapping

    def _bins(self, feature):
        return min(len(feature), self.max_bins)

    def _zoomed_feature(self, feature):
        extra = int(self.settings['zoom'] * len(feature) / 2)
        feature = copy(feature)
        feature.start -= extra
        feature.stop += extra
        return feature

    def ip_panel(self, ax, feature):
        bins = self._bins(feature)
        x, y = self.ip.local_coverage(feature, bins=bins,
                                          **self.local_coverage_kwargs)
        if isinstance(self.ip, _genomic_signal.BamSignal):
            y /= (self.ip.mapped_read_count() / 1e6)
        ax.fill_between(x, y, y2=0, **self.ip_style)
        ax.axis('tight')
        ax.spines['right'].set_color('None')
        ax.spines['top'].set_color('None')
        ax.xaxis.set_ticks_position('bottom')
        ax.yaxis.set_ticks_position('left')
        return feature

    def control_panel(self, ax, feature):
        bins = self._bins(feature)
        x, y = self.control.local_coverage(feature, bins=bins,
                                               **self.local_coverage_kwargs)
        if isinstance(self.control, _genomic_signal.BamSignal):
            y /= (self.control.mapped_read_count() / 1e6)
        ax.fill_between(x, y, y2=0, **self.control_style)
        ax.axis('tight')
        ax.spines['right'].set_color('None')
        ax.spines['top'].set_color('None')
        ax.xaxis.set_ticks_position('bottom')
        ax.yaxis.set_ticks_position('left')
        return feature

    def peaks_panel(self, ax, feature):
        hits = self.peaks.intersect([feature], u=True)
        track = Track(hits, **self.peaks_style)
        ax.add_collection(track)
        for x in (
            ax.get_yticklines() + ax.get_xticklines() + ax.get_yticklabels()
            + ax.get_xticklabels()
        ):
            x.set_visible(False)
        ax.set_ylabel('Peaks', rotation=0, horizontalalignment='right',
                      verticalalignment='center')
        ax.yaxis.set_ticks_position('none')
        ax.xaxis.set_ticks_position('none')
        ax.set_frame_on(False)

        return feature

    def plot(self, feature):
        self._current_feature = feature
        axes = super(ChIPSeqMiniBrowser, self).plot(feature)
        for ax in self.fig.axes:
            if ax is not self._last_ax:
                for txt in ax.get_xticklabels():
                    txt.set_visible(False)
        self.fig.subplots_adjust(wspace=0.05)
        self._first_ax.axis(xmin=feature.start, xmax=feature.stop)
        self._last_ax.set_xlabel(feature.chrom)
        self._last_ax.xaxis.set_major_formatter(FormatStrFormatter("%d"))
        return axes

    def coords(self):
        chrom = self._current_feature.chrom
        start, stop, _, _ = self._first_ax.axis()
        return '%s:%s-%s' % (chrom, int(start), int(stop))

    def genes_panel(self, ax, feature):
        """
        Plots gene models on an Axes.

        Queries the database

        :param ax: matplotlib.Axes object
        :param feature: pybedtools.Interval

        """
        from gffutils.contrib.plotting import Gene
        extent = [feature.start, feature.stop]
        nearby_genes = self.db.region(
            (feature.chrom, feature.start, feature.stop), featuretype='gene')
        ybase = 0
        ngenes = 0
        for nearby_gene in nearby_genes:
            # TODO: there should be a better way of identifying which gene is
            # the same as the feature requested.  Might need to expose an "ID"
            # kwarg.
            try:
                if nearby_gene['ID'][0] == feature['ID']:
                    color = '0.2'
                else:
                    color = '0.5'
            except KeyError:
                color = '0.5'
            ngenes += 1
            extent.extend([nearby_gene.start, nearby_gene.stop])
            gene_collection = Gene(
                self.db,
                nearby_gene,
                transcripts=None,
                cds=['CDS'],
                utrs=['exon'],
                ybase=ybase,
                color=color)
            gene_collection.name = nearby_gene.id
            gene_collection.add_to_ax(ax)
            ybase += gene_collection.max_y

        xmin = min(extent)
        xmax = max(extent)
        ymax = ngenes

        # 1% padding seems to work well
        padding = (xmax - xmin) * 0.01
        ax.axis('tight')

        # add lines indicating extent of current feature
        #vline_kwargs = dict(color='k', linestyle='--')
        #ax.axvline(feature.start, **vline_kwargs)
        #ax.axvline(feature.stop, **vline_kwargs)

        # Make a new feature to represent the region plus surrounding genes
        interval = pybedtools.create_interval_from_list(feature.fields)
        interval.strand = '.'
        for txt in ax.get_yticklabels():
            txt.set_visible(False)
        for tick in ax.get_yticklines():
            tick.set_visible(False)
        ax.set_ylabel('Genes')
        ax.spines['right'].set_color('None')
        ax.spines['left'].set_color('None')
        ax.spines['top'].set_color('None')
        ax.yaxis.set_ticks_position('none')
        ax.xaxis.set_ticks_position('bottom')
        ax.set_ylabel('Genes', rotation=0, horizontalalignment='right',
                      verticalalignment='center')

        return interval


class SignalMiniBrowser(BaseMiniBrowser):
    def __init__(self, genomic_signal_objs, local_coverage_kwargs=None,
                 plotting_kwargs=None):
        """
        Base class for plotting genomic signal.

        Plots genomic signal over a particular area of the genome.  Designed to
        be extended by subclasses, but can stand alone if all you want is
        a simple one-panel browser.

        :param genomic_signal_objs: list of genomic signal objects (e.g.,
            :class:`metaseq.genomic_signal.BamSignal` instances).
        :param local_coverage_kwargs: a dictionary of kwargs to send to each
            genomic signals' `local_coverage()` method, e.g.::

                local_coverage_kwargs = dict(fragment_size=200).
        :param plotting_kwargs:  a list of dictionaries, one for each genomic
            signals object, e.g,::

                plotting_kwargs = [dict(color='r', label='IP', dict(color='k',
                    label='input')]

        """
        super(SignalMiniBrowser, self).__init__(genomic_signal_objs)
        self.plotting_kwargs = \
            plotting_kwargs or [{} for i in genomic_signal_objs]
        self.local_coverage_kwargs = local_coverage_kwargs or {}

    def panels(self):
        """
        Add a single panel to the figure
        """
        ax = self.fig.add_subplot(111)
        return [(ax, self.signal_panel)]

    def signal_panel(self, ax, feature):
        """
        Plots each genomic signal as a line using the corresponding
        plotting_kwargs
        """
        for gs, kwargs in zip(self.genomic_signal_objs, self.plotting_kwargs):
            x, y = gs.local_coverage(feature, **self.local_coverage_kwargs)
            ax.plot(x, y, **kwargs)
        ax.axis('tight')
        return feature


class GeneModelMiniBrowser(SignalMiniBrowser):
    def __init__(self, genomic_signal_objs, db, **kwargs):
        """
        Mini-browser to show a signal panel on top and gene models on the
        bottom.

        :param genomic_signal_objs: a list of genomic_signals objects
        :param db: a `gffutils.FeatureDB`
        """
        super(GeneModelMiniBrowser, self).__init__(
            genomic_signal_objs, **kwargs)
        self.db = db

    def panels(self):
        """
        Add 2 panels to the figure, top for signal and bottom for gene models
        """
        ax1 = self.fig.add_subplot(211)
        ax2 = self.fig.add_subplot(212, sharex=ax1)
        return (ax2, self.gene_panel), (ax1, self.signal_panel)

    def gene_panel(self, ax, feature):
        """
        Plots gene models on an Axes.

        Queries the database

        :param ax: matplotlib.Axes object
        :param feature: pybedtools.Interval

        """
        from gffutils.contrib.plotting import Gene
        extent = [feature.start, feature.stop]
        nearby_genes = self.db.region(
            (feature.chrom, feature.start, feature.stop), featuretype='gene')
        ybase = 0
        ngenes = 0
        for nearby_gene in nearby_genes:
            ngenes += 1
            extent.extend([nearby_gene.start, nearby_gene.stop])
            gene_collection = Gene(
                self.db,
                nearby_gene,
                transcripts=['mRNA'],
                cds=['CDS'],
                utrs=['exon'],
                ybase=ybase,
                color="0.5", picker=5)
            gene_collection.name = nearby_gene.id
            gene_collection.add_to_ax(ax)
            ybase += gene_collection.max_y

        xmin = min(extent)
        xmax = max(extent)
        ymax = ngenes

        # 1% padding seems to work well
        padding = (xmax - xmin) * 0.01
        ax.axis('tight')

        # add lines indicating extent of current feature
        vline_kwargs = dict(color='k', linestyle='--')
        ax.axvline(feature.start, **vline_kwargs)
        ax.axvline(feature.stop, **vline_kwargs)

        # Make a new feature to represent the region plus surrounding genes
        interval = pybedtools.create_interval_from_list(feature.fields)
        interval.start = xmin - padding
        interval.stop = xmax + padding
        interval.strand = '.'
        return interval


class PeakMiniBrowser(SignalMiniBrowser):
    def __init__(self, genomic_signal_objs, bed, **kwargs):
        """
        Signal on the top panel, peaks in the bottom panel. `bed` is a filename
        or BedTool object (or list of these things) that will be used to make
        a pybedtools.contrib.plotting.Track.

        If genomic_signal_objs is None, then only show the peaks axes
        """
        super(PeakMiniBrowser, self).__init__(genomic_signal_objs, **kwargs)
        self.bed = bed

    def panels(self):
        ax1 = self.fig.add_subplot(211)
        ax2 = self.fig.add_subplot(212, sharex=ax1)
        return (ax1, self.signal_panel), (ax2, self.peak_panel)

    def peak_panel(self, ax, feature):
        bedtool = pybedtools.BedTool(self.bed)
        features = bedtool.intersect([feature], u=True)
        track = Track(features)
        ax.add_collection(track)
        #ax.axis('tight')
        return feature


if __name__ == "__main__":
    import metaseq
    import gffutils
    import pybedtools

    G = gffutils.FeatureDB(
        metaseq.example_filename('Homo_sapiens.GRCh37.66.cleaned.gtf.db'))

    ip = metaseq.genomic_signal(
        metaseq.example_filename('wgEncodeUwTfbsK562CtcfStdAlnRep1.bam'),
        'bam')
    inp = metaseq.genomic_signal(
        metaseq.example_filename('wgEncodeUwTfbsK562InputStdAlnRep1.bam'),
        'bam')
    peaks = pybedtools.BedTool(metaseq.example_filename(
        'wgEncodeUwTfbsK562CtcfStdPkRep1.narrowPeak.gz'))

    plotting_kwargs = [
        dict(color='r', label='IP'),
        dict(color='k', linestyle=':', label='input')]

    local_coverage_kwargs = dict(fragment_size=200)

    b = SignalMiniBrowser(
        [ip, inp],
        plotting_kwargs=plotting_kwargs,
        local_coverage_kwargs=local_coverage_kwargs)

    g = GeneModelMiniBrowser(
        [ip, inp],
        G,
        plotting_kwargs=plotting_kwargs,
        local_coverage_kwargs=local_coverage_kwargs)

    p = PeakMiniBrowser(
        [ip, inp],
        peaks,
        plotting_kwargs=plotting_kwargs,
        local_coverage_kwargs=local_coverage_kwargs)

    feature = peaks[3]

    b.plot(feature)
    g.plot(feature)
    p.plot(feature)
    plt.show()



"""
Many of these tests use the minimal test/data/gdc.bed file which has just
enough complexity to be useful in testing corner cases.  When reading through
the tests, it's useful to have that file open to understand what's happening.
"""
import os
import metaseq
import multiprocessing
from metaseq.array_helpers import ArgumentError
import numpy as np
from nose.tools import assert_raises
gs = {}
for kind in ['bed', 'bam', 'bigbed', 'bigwig']:
    gs[kind] = metaseq.genomic_signal(metaseq.example_filename('gdc.%s' % kind), kind)

PROCESSES = int(os.environ.get("METASEQ_PROCESSES", multiprocessing.cpu_count()))

def test_tointerval():
    assert metaseq.helpers.tointerval("chr2L:1-10[-]").strand == '-'
    assert metaseq.helpers.tointerval("chr2L:1-10[+]").strand == '+'
    assert metaseq.helpers.tointerval("chr2L:1-10").strand == '.'


def test_local_count():

    def check(kind, coord, expected, stranded):
        result = gs[kind].local_count(coord, stranded=stranded)
        assert result == expected, (kind, coord, result)

    for kind in ['bam', 'bigbed', 'bed']:
        for coord, expected, stranded in (
            ('chr2L:1-80', 3, False),       #  easy case
            ('chr2L:1000-3000', 0, False),  #  above upper boundary
            ('chr2L:1-9', 0, False),        #  below lower boundary
            ('chr2L:71-73[-]', 2, False),   #  unstranded = 2
            ('chr2L:71-73[-]', 1, True),    #  stranded = 1
            ('chr2L:70-71', 2, False),      #  pathological corner case
            ('chr2L:75-76', 0, False),      #  pathological corner case
        ):
            yield check, kind, coord, expected, stranded


def test_local_coverage_stranded():
    def check(kind, coord, expected):
        result = gs[kind].local_coverage(coord)
        assert np.all(result[0] == expected[0]) and np.all(result[1] == expected[1]), (kind, coord, result)

    for kind in ['bam', 'bigbed', 'bed', 'bigwig']:
        for coord, expected in (
            ('chr2L:1-20[-]',
             (
                 np.array([1,  2,  3,  4,  5,  6,  7,  8,  9,  10, 11, 12, 13, 14, 15, 16, 17, 18, 19]),
                 np.array([0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 1., 1., 1., 1., 0., 0., 0., 0., 0.])[::-1],
                 # note reverse------------------------------------------------------------------------^^^^^^
             ),
            ),
            ('chr2L:68-76[-]',
             (
                 np.array([68, 69, 70, 71, 72, 73, 74, 75]),
                 np.array([0., 0., 2., 2., 2., 2., 2., 0.])[::-1],
                 # note reverse----------------------------^^^^^^

             ),
            ),
        ):
            yield check, kind, coord, expected


def test_local_coverage_shifted():
    def check(kind, coord, shift_width, expected):
        result = gs[kind].local_coverage(coord, shift_width=shift_width)
        assert np.all(result[0] == expected[0]) and np.all(result[1] == expected[1]), (kind, coord, result)

    for kind in ['bam', 'bigbed', 'bed']:
        for coord, shift_width, expected in (
            ('chr2L:1-20', -2,
             (
                 np.array([1,  2,  3,  4,  5,  6,  7,  8,  9,  10, 11, 12, 13, 14, 15, 16, 17, 18, 19]),
                 np.array([0., 0., 0., 0., 0., 0., 0., 1., 1., 1., 1., 1., 0., 0., 0., 0., 0., 0., 0.]),
             ),
            ),
            # this one is complex, because the minus-strand read shifts left,
            # and the plus-strand shifts right.
            ('chr2L:68-76', 1,
             (
                 np.array([68, 69, 70, 71, 72, 73, 74, 75]),
                 np.array([0., 1., 1., 2., 2., 2., 1., 1.]),
             ),
            ),

            # shift the reads all the way out of the window...
            ('chr2L:68-76', 10,
             (
                 np.array([68, 69, 70, 71, 72, 73, 74, 75]),
                 np.array([0., 0., 0., 0., 0., 0., 0., 0.]),
             ),
            ),
        ):
            yield check, kind, coord, shift_width, expected


def test_local_coverage_read_strand():
    """
    checks stranded full binning

    excludes bigwig since strand doesn't make sense for that format.
    """
    def check(kind, coord, read_strand, expected):
        result = gs[kind].local_coverage(coord, read_strand=read_strand)
        assert np.all(result[0] == expected[0]) and np.all(result[1] == expected[1]), (kind, coord, result)

    for kind in ['bam', 'bigbed', 'bed']:
        for coord, read_strand, expected in (
            ('chr2L:1-20', '+',
             (
                 np.array([1,  2,  3,  4,  5,  6,  7,  8,  9,  10, 11, 12, 13, 14, 15, 16, 17, 18, 19]),
                 np.array([0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 1., 1., 1., 1., 0., 0., 0., 0., 0.]),
             ),
            ),
            ('chr2L:1-20', '-',
             (
                 np.array([1,  2,  3,  4,  5,  6,  7,  8,  9,  10, 11, 12, 13, 14, 15, 16, 17, 18, 19]),
                 np.array([0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.]),
             ),
            ),
        ):
            yield check, kind, coord, read_strand, expected


def test_local_coverage_fragment_size():
    def check(kind, coord, fragment_size, expected):
        result = gs[kind].local_coverage(coord, fragment_size=fragment_size)
        assert np.all(result[0] == expected[0]) and np.all(result[1] == expected[1]), (kind, coord, result)

    for kind in ['bam', 'bigbed', 'bed']:
        for coord, fragment_size, expected in (
            ('chr2L:1-20', 7,
             (
                 np.array([1,  2,  3,  4,  5,  6,  7,  8,  9,  10, 11, 12, 13, 14, 15, 16, 17, 18, 19]),
                 np.array([0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 1., 1., 1., 1., 1., 1., 0., 0., 0.]),
             ),
            ),

            ('chr2L:68-76', 6,
             (
                 np.array([68, 69, 70, 71, 72, 73, 74, 75]),
                 np.array([0., 1., 2., 2., 2., 2., 2., 1.]),
             ),
            ),

            ('chr2L:68-76', 1,
             (
                 np.array([68, 69, 70, 71, 72, 73, 74, 75]),
                 np.array([0., 0., 1., 0., 0., 0., 1., 0.]),
             ),
            ),
        ):
            yield check, kind, coord, fragment_size, expected


def test_local_coverage_score():
    def check(kind, coord, expected):
        result = gs[kind].local_coverage(coord, use_score=True)
        assert np.all(result[0] == expected[0]) and np.all(result[1] == expected[1]), (kind, coord, result)

    for kind in ['bigbed', 'bed']:
        for coord, expected in (
            ('chr2L:1-20',
             (
                 np.array([1,  2,  3,  4,  5,  6,  7,  8,  9,  10,   11,   12,   13,   14,   15, 16, 17, 18, 19]),
                 np.array([0., 0., 0., 0., 0., 0., 0., 0., 0., 255., 255., 255., 255., 255., 0., 0., 0., 0., 0.]),
             ),
            ),
            ('chr2L:68-76',
             (
                 np.array([68, 69, 70,   71,   72,   73,   74,   75]),
                 np.array([0., 0., 510., 510., 510., 510., 510., 0.]),
             ),
            ),
        ):
            yield check, kind, coord, expected


def test_local_coverage_full():
    """generator of tests for local coverage

    ensures that all formats are consistent in their results when retrieving
    the full un-binned data.
    """
    def check(kind, coord, processes, expected):
        result = gs[kind].local_coverage(coord, processes=processes)
        assert np.all(result[0] == expected[0]) and np.all(result[1] == expected[1]), (kind, coord, result)

    for kind in ['bam', 'bigbed', 'bed', 'bigwig']:
        for coord, expected in (
            ('chr2L:1-20',
             (
                 np.array([1,  2,  3,  4,  5,  6,  7,  8,  9,  10, 11, 12, 13, 14, 15, 16, 17, 18, 19]),
                 np.array([0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 1., 1., 1., 1., 0., 0., 0., 0., 0.]),
             ),
            ),
            ('chr2L:68-76',
             (
                 np.array([68, 69, 70, 71, 72, 73, 74, 75]),
                 np.array([0., 0., 2., 2., 2., 2., 2., 0.]),
             ),
            ),
            ('chr2L:568-576',
             (
                 np.array([568, 569, 570, 571, 572, 573, 574, 575]),
                 np.array([0., 0.,   0.,  0.,  0.,  0.,  0.,  0.]),
             ),
            ),
        ):
            for processes in [None, PROCESSES]:

                yield check, kind, coord, processes, expected


def test_local_coverage_binned():
    """generator of tests for local coverage

    ensures that all formats are consistent in their results when retrieving
    binned data.
    """
    def check(kind, coord, processes, expected):
        if kind == 'bigwig':
            result = gs[kind].local_coverage(coord, bins=8, method='get_as_array', processes=processes)
        else:
            result = gs[kind].local_coverage(coord, bins=8, processes=processes)
        try:
            assert np.allclose(result[0], expected[0]) and np.allclose(result[1], expected[1])
        except:
            print (kind, coord, result, expected)
            raise

    for kind in ['bam', 'bigbed', 'bed', 'bigwig']:
        for coord, expected in (
            ('chr2L:1-20',
             (
                 np.array([ 1., 3.57142857, 6.14285714, 8.71428571, 11.28571429, 13.85714286, 16.42857143, 19.]),
                 np.array([ 0., 0.,         0.,         0.,         1.,          1.,          0.,          0. ]),
             ),
            ),
            ('chr2L:68-76',
             (
                 np.array([68, 69, 70, 71, 72, 73, 74, 75]),
                 np.array([0., 0., 2., 2., 2., 2., 2., 0.]),
             ),
            ),
        ):
            for processes in [None, PROCESSES]:
                yield check, kind, coord, processes, expected


def test_array_binned():
    def check(kind, coord, processes, expected):
        if kind == 'bigwig':
            result = gs[kind].array(coord, bins=8, method='get_as_array', processes=processes)
        else:
            result = gs[kind].array(coord, bins=8, processes=processes)
        try:
            assert np.allclose(result, expected)
        except:
            print (kind, coord, result, expected)
            raise

    for kind in ['bam', 'bigbed', 'bed', 'bigwig']:
        for coord, expected in (
            (['chr2L:1-20'],
             np.array([[0., 0., 0., 0., 1., 1., 0., 0. ]]),
            ),
            (['chr2L:1-20', 'chr2L:1-20[-]'],
             np.array([[0., 0., 0., 0., 1., 1., 0., 0. ],
                       [0., 0., 1., 1., 0., 0., 0., 0. ]]),
            ),
            (['chr2L:68-76'],
             np.array([[0., 0., 2., 2., 2., 2., 2., 0.]]),
             ),
        ):
            for processes in [None, PROCESSES]:
                yield check, kind, coord, processes, expected


def test_array_binned_preserve_total():
    def check(kind, coord, processes, expected):
        kwargs = dict(features=coord, bins=8, processes=processes, preserve_total=True)
        if kind == 'bigwig':
            assert_raises(ArgumentError, gs[kind].array, method='get_as_array', **kwargs)
            return
        else:
            result = gs[kind].array(**kwargs)
        try:
            assert np.allclose(result, expected)
        except:
            print (kind, coord, result, expected)
            raise

    for kind in ['bam', 'bigbed', 'bed', 'bigwig']:
        for coord, expected in (
            (['chr2L:1-20'],
             np.array([[0., 0., 0., 0., .5, .5, 0., 0. ]]),
            ),
            (['chr2L:1-20', 'chr2L:1-20[-]'],
             np.array([[0., 0., 0., 0., .5, .5, 0., 0. ],
                       [0., 0., .5, .5, 0., 0., 0., 0. ]]),
            ),
            (['chr2L:68-76'],
             np.array([[0., 0., .4, .4, .4, .4, .4, 0.]]),
             ),
        ):
            for processes in [None, PROCESSES]:
                yield check, kind, coord, processes, expected


def test_invalid_arguments():
    def check(kind, kw):
        assert_raises(ArgumentError, gs[kind].array, **kw)

    default_kwargs = dict(features='chr2L:1-20', bins=8)
    # List the kinds of args that should return ArgumentErrors.
    invalid_kwargs = {
        'bam': (
            dict(use_score=True),
        ),

        'bigwig': (
            dict(read_strand=True),
            dict(fragment_size=100),
            dict(shift_width=100), 
            dict(use_score=True),
            dict(preserve_total=True),
        ),
    }
    for kind, kwarg_list in sorted(invalid_kwargs.items()):
        for kwargs in kwarg_list:
            default_kwargs_copy = default_kwargs.copy()
            default_kwargs_copy.update(kwargs)
            yield check, kind, default_kwargs_copy


def test_array_ragged():
    def check(kind, coord, processes, expected):
        if kind == 'bigwig':
            result = gs[kind].array(coord, method='get_as_array', ragged=True, processes=processes)
        else:
            result = gs[kind].array(coord, processes=processes, ragged=True)
        try:
            if isinstance(result, list):
                for i, j in zip(result, expected):
                    assert np.allclose(i, j)
        except:
            print (kind, coord, result, expected)
            raise

    for kind in ['bam', 'bigbed', 'bed', 'bigwig']:
        for coord, expected in (
            (['chr2L:1-20'],
             np.array([[0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  1.,  1.,  1.,  1.,   1.,  0.,  0.,  0.,  0.,  0.]])
            ),
            (
                [['chr2L:1-20'], ['chr2L:1-19']],
                [
                    np.array([ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  1.,  1.,  1.,  1.,   1.,  0.,  0.,  0.,  0.,  0.]),
                    np.array([ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  1.,  1.,  1.,  1.,   1.,  0.,  0.,  0.,  0., ])
                ]
            ),

            ([['chr2L:68-76', 'chr2L:68-76'], 'chr2L:68-76'],
             [
                 np.array([0., 0., 2., 2., 2., 2., 2., 0., 0., 0., 2., 2., 2., 2., 2., 0.]),
                 np.array([0., 0., 2., 2., 2., 2., 2., 0.])
             ],
            ),
        ):
            for processes in [None, PROCESSES]:
                yield check, kind, coord, processes, expected


def test_bigwig_methods():
    """
    using bx-python's `summarize` gives different results than UCSC's summarize
    and summarizing from the original BAM.

    possible bx-python bug, or just a resolution thing?  Haven't worked that
    out yet, so test results here.
    """
    # y0 is identical to the above test, and should be the same as
    # local_coverage for bam, bigbed, and bed as well.
    #
    # y1 may not always be identical . . . I think it's due to resolution of
    # the summarized file, especially when working with such short genomic
    # intervals?
    x0 = np.array([1., 3.57142857, 6.14285714, 8.71428571, 11.28571429, 13.85714286, 16.42857143, 19.])
    y0 = np.array([ 0., 0.,         0.,         0.,         1.,          1.,          0.,          0. ]),
    y1 = np.array([ 0., 0.,         0.,         0.,         1.,          1.,          1.,          0. ]),
    #                                                         note difference here ---^
    x, y = gs['bigwig'].local_coverage('chr2L:1-20', bins=8, method='get_as_array')
    assert np.allclose(x0, x)
    assert np.allclose(y0, y)

    x, y = gs['bigwig'].local_coverage('chr2L:1-20', bins=8, method='summarize')
    assert np.allclose(x0, x)
    assert np.allclose(y1, y)

    x, y = gs['bigwig'].local_coverage('chr2L:1-20', bins=8, method='ucsc_summarize')
    assert np.allclose(x0, x)
    assert np.allclose(y0, y)


def test_nonbigwig_kwargs():
    """
    test non-bigwig kwargs with a bigwig file

    these kwargs are not valid for bigwigs because they don't make sense for
    the format.
    """
    nonwigs = {
        'read_strand': '+',
        'fragment_size': 200,
        'shift_width': 80,
        'use_score': True,
    }
    for nw in nonwigs.items():
        kwargs = dict((nw,))
        try:
            assert_raises(
                ArgumentError, gs['bigwig'].local_coverage, 'chr2L:1-20',
                **kwargs)
        except AssertionError:
            print kwargs
            raise

    assert_raises(ArgumentError, gs['bigwig'].local_coverage, 'chr2L:1-20',
                  method='ucsc_summarize', preserve_total=True)


def test_local_coverage_kwarg_errors():
    """
    ensures the argument-checking is working
    """
    def check(kind, kwargs):
        assert_raises(
            ArgumentError,
            gs[kind].local_coverage, **kwargs)

    items = [
        ('bam', dict(bins='a', features='chr2L:1-20')),
        ('bam', dict(use_score=True, features='chr2L:1-20')),
        ('bam', dict(bins=[8, 10], features=['chr2L:1-20'])),

    ]

    for kind, kwargs in items:
        yield check, kind, kwargs


def test_local_coverage_full_multifeature():
    def check(kind, coord, expected):
        result = gs[kind].local_coverage(coord)
        assert np.all(result[0] == expected[0]) and np.all(result[1] == expected[1]), (kind, coord, result)

    for kind in ['bam', 'bigbed', 'bed', 'bigwig']:
        for coord, expected in (
            (['chr2L:1-20', 'chr2L:68-76'],
             (
                 np.array([1,  2,  3,  4,  5,  6,  7,  8,  9,  10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 68, 69, 70, 71, 72, 73, 74, 75]),
                 np.array([0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 1., 1., 1., 1., 0., 0., 0., 0., 0., 0., 0., 2., 2., 2., 2., 2., 0.]),
             ),
            ),
        ):
            yield check, kind, coord, expected


def test_errors():
    "these things should raise errors"
    def check(error, callable_obj, args, kwargs):
        assert_raises(error, callable_obj, *args, **kwargs)

    class X(metaseq.filetype_adapters.BaseAdapter):
        def make_fileobj(self):
            return None

    items = [
        (ValueError, metaseq.filetype_adapters.BaseAdapter, (metaseq.example_filename('gdc.bed'),), {}),
        (NotImplementedError, metaseq.filetype_adapters.BigWigAdapter(metaseq.example_filename('gdc.bigwig')).__getitem__, (0,), {}),
        (ValueError, X("").__getitem__, (0,), {}),
        #(ValueError, gs['bam'].local_coverage, ['chr2L:1-5', 'chr2L:1-5'], dict(processes=PROCESSES)),
    ]
    for error, callable_obj, args, kwargs in items:
        yield check, error, callable_obj, args, kwargs

def test_supported_formats():
    assert set(metaseq._genomic_signal.supported_formats()) \
        == set(['bam', 'bigwig', 'bed', 'gff', 'gtf', 'vcf', 'bigbed'])
    assert_raises(ValueError, metaseq.genomic_signal, "", 'unsupported')

def test_bam_genome():
    assert gs['bam'].genome() == {'chr2L': (0, 23011544L)}

def test_bam_mmr():
    import os
    assert gs['bam'].mapped_read_count(force=True) == 8
    gs['bam']._readcount = None
    assert gs['bam'].mapped_read_count(force=False) == 8, gs['bam'].mapped_read_count(force=False)

def test_bigwig_out_of_range():
    x, y = gs['bigwig'].local_coverage('chr1:1-100', bins=None)
    assert y.sum() == 0
    x, y = gs['bigwig'].local_coverage('chr1:1-100', bins=None, method='ucsc_summarize')
    assert y.sum() == 0

    x, y = gs['bigwig'].local_coverage('chr2L:1-100', bins=None, method='ucsc_summarize')





"""
This module integrates parts of metaseq that are useful for ChIP-seq analysis.
"""
import os
from itertools import izip
import gffutils
from gffutils.helpers import asinterval
import metaseq
import pybedtools
from metaseq.minibrowser import SignalMiniBrowser, GeneModelMiniBrowser
import numpy as np
from matplotlib import pyplot as plt
import matplotlib
import yaml

def save(c, prefix, relative_paths=True):
    """
    Save data from a Chipseq object.


    Parameters
    ----------
    c : Chipseq object
        Chipseq object, most likely after calling the `diffed_array` method

    prefix : str
        Prefix, including any leading directory paths, to save the data.

    relative_paths : bool
        If True (default), then the path names in the `prefix.info` file will
        be relative to `prefix`.  Otherwise, they will be absolute.

    The following files will be created:

    :prefix.intervals:
        A BED file (or GFF, GTF, or VCF as appropriate) of the features used for the array

    :prefix.info:
        A YAML-format file indicating the IP and control BAM files, any array
        kwargs, the database filename, and any minibrowser local coverage args.
        These are all needed to reconstruct a new Chipseq object. Path names
        will be relative to `prefix`.

    :prefix.npz:
        A NumPy .npz file with keys 'diffed_array', 'ip_array', and 'control_array'

    """
    dirname = os.path.dirname(prefix)

    pybedtools.BedTool(c.features).saveas(prefix + '.intervals')

    def usepath(f):
        if relative_paths:
            return os.path.relpath(f, start=dirname)
        else:
            return os.path.abspath(f)

    with open(prefix + '.info', 'w') as fout:
        info = {
            'ip_bam': usepath(c.ip.fn),
            'control_bam': usepath(c.control.fn),
            'array_kwargs': c.array_kwargs,
            'dbfn': usepath(c.dbfn),
            'browser_local_coverage_kwargs': c.browser_local_coverage_kwargs,
            'relative_paths': relative_paths,
        }
        fout.write(yaml.dump(info, default_flow_style=False))
    np.savez(
        prefix,
        diffed_array=c.diffed_array,
        ip_array=c.ip_array,
        control_array=c.control_array
    )

def load(prefix):
    info = yaml.load(open(prefix + '.info'))
    c = Chipseq(ip_bam=info['ip_bam'], control_bam=info['control_bam'],
                dbfn=info['dbfn'])
    npz = np.load(prefix + '.npz', mmap_mode='r')
    c.ip_array = npz['ip_array']
    c.control_array = npz['control_array']
    c.diffed_array = npz['diffed_array']
    c.array_kwargs = info['array_kwargs']
    c.features = pybedtools.BedTool(prefix + '.intervals')
    c.browser_local_coverage_kwargs = info['browser_local_coverage_kwargs']
    return c

class GeneChipseqMiniBrowser(GeneModelMiniBrowser):
    def __init__(self, genomic_signal_objs, db, **kwargs):
        super(GeneChipseqMiniBrowser, self).__init__(
            genomic_signal_objs, db, **kwargs)

    def plot(self, feature):
        super(GeneChipseqMiniBrowser, self).plot(feature)
        ax1, ax2 = self.fig.axes
        ax1.legend(loc='best')
        ax1.set_ylabel('RPMMR')
        ax1.xaxis.set_visible(False)
        ax2.xaxis.set_major_formatter(
            matplotlib.ticker.FormatStrFormatter('%d'))
        ax2.set_xlabel(feature.chrom)


class SignalChipseqMiniBrowser(SignalMiniBrowser):
    def __init__(self, genomic_signal_objs, **kwargs):
        super(SignalChipseqMiniBrowser, self).__init__(
            genomic_signal_objs, **kwargs)

    def plot(self, feature):
        super(SignalChipseqMiniBrowser, self).plot(feature)
        ax1, = self.fig.axes
        ax1.legend(loc='best')
        ax1.set_ylabel('RPMMR')
        ax1.set_xlabel(feature.chrom)


class Chipseq(object):
    """
    Class for visualizing and interactively exploring ChIP-seq data.

    Needs two BAM files (for IP and control) and a gffutils database
    filename in order to display gene models.

    Typical usage is to create a normalized array of signal over each
    feature with the `diff_array` method, and then plot with the `plot`
    method.

    The resulting figure has the matrix as a heatmap, the average signal
    over features, and a panel with points that can be zoomed and clicked,
    spawning a mini-browser window for the corresponding feature.

    Configuration can be done by adjusting the following attributes after
    creating a Chipseq instance:

        * `_strip_kwargs` (the style for the dots in the left panel)
        * `browser_plotting_kwargs` (style of signal lines in the mini-browser)

    """
    def __init__(self, ip_bam, control_bam, dbfn=None):
        """
        Set up a :class:`Chipseq` object.


        :param ip_bam: filename of BAM file for ChIP data
        :param control_bam: filename of BAM file for control data
        :param dbfn: filename of gffutils database
        """
        self.ip = metaseq.genomic_signal(ip_bam, kind='bam')
        self.control = metaseq.genomic_signal(control_bam, kind='bam')
        self.dbfn = dbfn
        self.db = None
        if self.dbfn:
            self.db = gffutils.FeatureDB(dbfn)
        self.ip_array = None
        self.control_array = None

        self._strip_kwargs = dict(
            color='.5', markeredgewidth=0, marker='o', linestyle='None',
            picker=5)
        self.browser_plotting_kwargs = [
            dict(color='r', label='IP'),
            dict(color='k', linestyle=':', label='control')
        ]


    def diff_array(self, features, force=True, func=None,
                   array_kwargs=dict(), cache=None):
        """
        Scales the control and IP data to million mapped reads, then subtracts
        scaled control from scaled IP, applies `func(diffed)` to the diffed
        array, and finally sets `self.diffed_array` to be the result.

        Arrays `self.ip` and `self.control` are set as well, and if
        `force=False`, then previously-created arrays will be used instead of
        re-calculating new ones.  This is useful if you want to easily try
        multiple `func` functions without having to re-calculate the data.

        Another side-effect is that `self.features` is set so that it can be
        accesed by other methods.

        :param features: a list of pybedtools.Interval objects
        :param array_kwargs: extra keyword args passed to genomic_signal.array;
            typically this will include `bins`, `processes`, and `chunksize`
            arguments.
        :param func: a function to apply to the diffed arrays. By default
            this is :func:`metaseq.plotutils.nice_log`; another option might be
            `lambda x: x`, or `lambda x: 1e6*x`
        :param force: Force a re-calculation of the arrays; otherwise uses
            cached values
        """
        self.features = list(features)
        self.browser_local_coverage_kwargs = array_kwargs.copy()
        self.browser_local_coverage_kwargs.pop('processes', None)
        self.browser_local_coverage_kwargs.pop('chunksize', 1)

        self.array_kwargs = array_kwargs.copy()

        if (self.ip_array is None) or force:
            self.ip_array = self.ip.array(features, **array_kwargs)
            self.ip_array /= self.ip.mapped_read_count() / 1e6

        if (self.control_array is None) or force:
            self.control_array = self.control.array(features, **array_kwargs)
            self.control_array /= self.control.mapped_read_count() / 1e6

        if func is None:
            #func = metaseq.plotutils.nice_log
            self.diffed_array = self.ip_array - self.control_array
        else:
            self.diffed_array = func(self.ip_array - self.control_array)

    def plot(self, x, row_order=None, imshow_kwargs=None, strip=True):
        """
        Plot the scaled ChIP-seq data.

        :param x: X-axis to use (e.g, for TSS +/- 1kb with 100 bins, this would
            be `np.linspace(-1000, 1000, 100)`)
        :param row_order: Array-like object containing row order -- typically
            the result of an `np.argsort` call.
        :param strip: Include axes along the left side with points that can be
            clicked to spawn a minibrowser for that feature.
        """
        nrows = self.diffed_array.shape[0]
        if row_order is None:
            row_order = np.arange(nrows)
        extent = (min(x), max(x), 0, nrows)
        axes_info = metaseq.plotutils.matrix_and_line_shell(strip=strip)
        fig, matrix_ax, line_ax, strip_ax, cbar_ax = axes_info
        _imshow_kwargs = dict(
            aspect='auto', extent=extent, interpolation='nearest')
        if imshow_kwargs:
            _imshow_kwargs.update(imshow_kwargs)

        if 'cmap' not in _imshow_kwargs:
            _imshow_kwargs['cmap'] = metaseq.colormap_adjust.smart_colormap(
                self.diffed_array.min(),
                self.diffed_array.max()
            )
        mappable = matrix_ax.imshow(
            self.diffed_array[row_order],
            **_imshow_kwargs)
        plt.colorbar(mappable, cbar_ax)
        line_ax.plot(x, self.diffed_array.mean(axis=0))
        if strip_ax:
            line, = strip_ax.plot(np.zeros((nrows,)), np.arange(nrows) + 0.5,
                                  **self._strip_kwargs)
            line.features = self.features
            line.ind = row_order

        matrix_ax.axis('tight')
        if strip_ax:
            strip_ax.xaxis.set_visible(False)
            matrix_ax.yaxis.set_visible(False)

        matrix_ax.xaxis.set_visible(False)

        if self.db:
            self.minibrowser = GeneChipseqMiniBrowser(
                [self.ip, self.control],
                db=self.db,
                plotting_kwargs=self.browser_plotting_kwargs,
                local_coverage_kwargs=self.browser_local_coverage_kwargs)
        else:
            self.minibrowser = SignalChipseqMiniBrowser(
                [self.ip, self.control],
                plotting_kwargs=self.browser_plotting_kwargs,
                local_coverage_kwargs=self.browser_local_coverage_kwargs)

        fig.canvas.mpl_connect('pick_event', self.callback)

        self.fig = fig
        self.axes = {
            'matrix_ax': matrix_ax,
            'strip_ax': strip_ax,
            'line_ax': line_ax,
            'cbar_ax': cbar_ax
        }

    def callback(self, event):
        """
        Callback function to spawn a mini-browser when a feature is clicked.
        """
        artist = event.artist
        ind = artist.ind
        limit = 5
        browser = True
        if len(event.ind) > limit:
            print "more than %s genes selected; not spawning browsers" % limit
            browser = False
        for i in event.ind:
            feature = artist.features[ind[i]]
            print feature,
            if browser:
                self.minibrowser.plot(feature)


def estimate_shift(signal, genome=None, windowsize=5000, thresh=None,
                   nwindows=1000, maxlag=500, array_kwargs=None,
                   verbose=False):
    """
    Experimental: cross-correlation to estimate the shift width of ChIP-seq
    data

    This can be interpreted as the binding site footprint.

    For ChIP-seq, the plus and minus strand reads tend to be shifted in the 5'
    direction away from each other.  Various ChIP-seq peak-callers estimate
    this distance; this function provides a quick, tunable way to do so using
    cross-correlation.  The resulting shift can then be incorporated into
    subsequent calls to `array` by adding the shift_width kwarg.


    :param signal: genomic_signal object
    :param genome: String assembly for constructing windows
    :param nwindows: Number of windows to compute cross-correlation on
    :param windowsize: Size of each window to compute cross-correlation on.
    :param thresh: Threshold read coverage to run cross-correlation on.  This
        is likely to be a function of the fragment size provided in
        `array_kwargs` `windowsize`.  If `thresh` is small, then the cross
        correlation can be noisy.
    :param maxlag: Max shift to look for
    :param array_kwargs: Kwargs passed directly to genomic_signal.array, with
        the default of `bins=windowsize` for single-bp resolution, and
        `read_strand` will be overwritten.
    :param verbose: Be verbose.

    Returns lags and a `maxlag*2+1` x `nwindows` matrix of cross-correlations.

    You can then plot the average cross-correlation function with::

        plt.plot(lags, shift.mean(axis=0))

    and get the distance to shift with::

        d = lags[np.argmax(shift.mean(axis=0))]

    and then plot that with::

        plt.axvline(d, color='k', linestyle='--')

    The number of windows with at least `thresh` coverage is::

        shift.shape[0]
    """
    if thresh is None:
        thresh = 0

    if genome is None:
        genome = signal.genome()

    if array_kwargs is None:
        array_kwargs = {}

    array_kwargs.pop('read_strand', None)

    if 'bins' not in array_kwargs:
        array_kwargs['bins'] = windowsize

    def add_strand(f, strand):
        fields = f.fields[:]
        while len(fields) < 5:
            fields.append('.')
        fields.append(strand)
        return pybedtools.create_interval_from_list(fields)

    windows = pybedtools.BedTool()\
        .window_maker(genome=genome, w=windowsize)

    random_subset = pybedtools.BedTool(windows[:nwindows])\
        .shuffle(genome=genome).saveas()

    if verbose:
        sys.stderr.write("Getting plus-strand signal for %s regions...\n"
                         % nwindows)
        sys.stderr.flush()

    plus = signal.array(
        features=random_subset,
        read_strand="+",
        **array_kwargs).astype(float)

    if verbose:
        sys.stderr.write("Getting minus-strand signal for %s regions...\n"
                         % nwindows)
        sys.stderr.flush()

    minus = signal.array(
        features=random_subset,
        read_strand="-",
        **array_kwargs).astype(float)

    # only do cross-correlation if you have enough reads to do so
    enough = ((plus.sum(axis=1) / windowsize) > thresh) \
        & ((minus.sum(axis=1) / windowsize) > thresh)

    if verbose:
        sys.stderr.write(
            "Running cross-correlation on %s regions that passed "
            "threshold\n" % sum(enough))
    results = np.zeros((sum(enough), 2 * maxlag + 1))
    for i, xy in enumerate(izip(plus[enough], minus[enough])):
        x, y = xy
        results[i] = xcorr(x, y, maxlag)

    lags = np.arange(-maxlag, maxlag + 1)

    return lags, results


def xcorr(x, y, maxlags):
    """
    Streamlined version of matplotlib's `xcorr`, without the plots.

    :param x, y: NumPy arrays to cross-correlate
    :param maxlags: Max number of lags; result will be `2*maxlags+1` in length
    """
    xlen = len(x)
    ylen = len(y)
    assert xlen == ylen

    c = np.correlate(x, y, mode=2)

    # normalize
    c /= np.sqrt(np.dot(x, x) * np.dot(y, y))

    lags = np.arange(-maxlags, maxlags + 1)
    c = c[xlen - 1 - maxlags:xlen + maxlags]

    return c


if __name__ == "__main__":
    import sys
    choices = ['xcorr', 'chipseq']
    try:
        examples = sys.argv[1:]
    except IndexError:
        print 'Choices are: ', choices
        examples = []

    for ex in examples:
        if ex not in choices:
            raise ValueError('%s not in %s' % (ex, choices))

    if 'xcorr' in examples:
        ip = metaseq.genomic_signal(
            metaseq.example_filename(
                'wgEncodeUwTfbsK562CtcfStdAlnRep1.bam'), 'bam')

        NWINDOWS = 5000
        FRAGMENT_SIZE = 1
        WINDOWSIZE = 5000
        THRESH = FRAGMENT_SIZE / float(WINDOWSIZE) * 10
        lags, shift = estimate_shift(
            ip, nwindows=NWINDOWS, maxlag=500, thresh=THRESH,
            array_kwargs=dict(
                processes=8, chunksize=100,
                fragment_size=FRAGMENT_SIZE),
            verbose=True)
        plt.plot(lags, shift.mean(axis=0))
        plt.axvline(
            lags[np.argmax(shift.mean(axis=0))],
            linestyle='--', color='k')

    if 'chipseq' in examples:
        # Example files...
        dbfn = metaseq.example_filename(
            'Homo_sapiens.GRCh37.66.cleaned.gtf.db')
        C = Chipseq(
            ip_bam=metaseq.example_filename(
                'wgEncodeUwTfbsK562CtcfStdAlnRep1.bam'),
            control_bam=metaseq.example_filename(
                'wgEncodeUwTfbsK562InputStdAlnRep1.bam'),
            dbfn=dbfn)

        # Make some features to use (TSS +/- 1kb)
        def generator():
            G = gffutils.FeatureDB(dbfn)
            genes = G.features_of_type('gene')
            for i in range(5000):
                yield asinterval(genes.next())

        from pybedtools.featurefuncs import TSS
        features = pybedtools.BedTool(generator())\
            .each(TSS, upstream=1000, downstream=1000)\
            .saveas()

        # x-axis for plots
        x = np.linspace(-500, 500, 100)

        # Create the array
        C.diff_array(
            features=features,
            array_kwargs=dict(
                fragment_size=200, bins=100, chunksize=50, processes=6))

        # sort genes by TIP zscore
        row_order = np.argsort(
            metaseq.plotutils.tip_zscores(C.diffed_array))[::-1]

        # Plot 'em using a nice red-to-blue colormap
        from metaseq.colormap_adjust import smart_colormap
        cmap = smart_colormap(C.diffed_array.min(), C.diffed_array.max())

        C.plot(x=x, row_order=row_order, imshow_kwargs=dict(cmap=cmap))
    plt.show()



#!/usr/bin/env python
usage="""
This is a benchmarking script for the metaseq Python package
(https://github.com/daler/metaseq).  It can help decide how many processors are
optimal to use for your particular hardware configuration.

The default arguments work for data shipped along with metaseq.  The default
arguments generate 1000 random features of about 5kb in size and then extract
signal data for BED, BAM, bigWig, and bigBed formats. This is done using from
1 to however many CPUs are availble.

The result is a plot showing the actual time in seconds per 1000 features, over
a range of processors.  A dotted line for the optimal 1/x speedup.  The plot is
also saved as a PDF.

Note that results are a function of CPU speed, disk I/O throughput, and nature
of the data (for BAM, BED, and bigBed formats), so your mileage may vary.  For
example, multiple CPUs will not help if there's only a small number of
features.
"""
import sys
import numpy as np
import time
from matplotlib import pyplot as plt
import pybedtools
import metaseq
import multiprocessing
import argparse
import os

ap = argparse.ArgumentParser(usage=usage)
ap.add_argument(
    '-n', '--nfeatures', type=int, default=1000,
    help='Number of random windows to make.  The size of each feature is '
    '(start - stop) / nfeatures, so changing this parameter will affect '
    'the size of each feature.')
ap.add_argument(
    '-c', '--chunksize', type=int, default=100,
    help='Number of windows to send to each process')
ap.add_argument(
    '--start', type=int, default=10000,
    help='start coord of possible windows. See --nfeatures help regarding size.')
ap.add_argument(
    '--stop', type=int, default=5000000,
    help='stop coord of possible windows. See --nfeatures help regarding size.')
ap.add_argument(
    '--chrom', default='chr2L',
    help='chromsome to make windows on')
ap.add_argument(
    '--type', default='all',
    help='Only use the specified file types.  Either "all" '
    '(default) or a comma-separated list of [bam, bigwig, bed, '
    'bigbed].')
ap.add_argument(
    '--prefix', default=os.path.join(metaseq.data_dir(), 'x'),
    help='Prefix of filenames to use.  Expects files '
    'with this prefix, and the following suffixes: .bam, .bigwig, '
    '.bed.gz (should already be tabixed), .bigbed.  Default: %(default)s')
ap.add_argument(
    '--plot-prefix', default='./speedtest',
    help='Filename used to save the resulting plot. Default is %(default)s')
ap.add_argument(
    '--bins', default=100,
    help='Number of bins for each feature')
args = ap.parse_args()


requested = args.type.split(',')
allowed = ['bam', 'bed', 'bigwig', 'bigbed', 'all']
for req in requested:
    if req not in allowed:
        raise ValueError("%s not in %s" % (req, allowed))

plt.rcParams['font.size'] = 10

intervals = pybedtools.BedTool().window_maker(
    genome={args.chrom: (args.start, args.stop)}, n=args.nfeatures)\
    .shuffle(genome={args.chrom: (args.start, args.stop)}, seed=1)

size = (args.stop - args.start) / args.nfeatures
if args.type == 'all':
    requested = ['bigwig', 'bam', 'bigbed', 'bed']
signals = []
for req in requested:
    if req != 'bed':
        signals.append(metaseq.genomic_signal(args.prefix + '.' + req, req))
    else:
        signals.append(metaseq.genomic_signal(args.prefix + '.bed.gz', 'bed'))

files = '* ' + '\n  * '.join([i.fn for i in signals])
plot_filename = (args.plot_prefix + '-%s_features-%s_bp_chunksize=%s.pdf'
                 % (args.nfeatures, size, args.chunksize))
processes = range(1, multiprocessing.cpu_count() + 1)
max_proc = processes[-1]
print """
{usage}
Parameters
----------
This script will generate {args.nfeatures} random features, each about {size}
bp, from genomic coordinates {args.chrom}:{args.start}-{args.stop}.

Genomic signal will be extracted from these files:
  {files}

For each file, the signal for the {args.nfeatures} features will be extracted
in parallel and binned into {args.bins} bins, using from 1 to {max_proc} CPUs.
Each CPU will get as many as {args.chunksize} features at a time.

Plot will be saved as {plot_filename}.
""".format(**locals())



d = {}
results = {}
for x in signals:
    print '\n' + x.kind, '[%s]' % x.fn
    sys.stdout.flush()
    times = []
    print 'CPUs:',
    for p in processes:
        print p,
        sys.stdout.flush()
        if p == 1:
            p = None
        t = []
        t0 = time.time()
        a = x.array(intervals, bins=args.bins, processes=p,
                    chunksize=args.chunksize)
        elapsed = time.time() - t0
        times.append(elapsed)
    results[x.kind] = a.sum(axis=0)
    d[x.kind] = times

fig = plt.figure(figsize=(12, 6))
i = 1
for k, t in d.items():
    ax = fig.add_subplot(2, len(signals), i)
    t = np.array(t) / (len(intervals) / 1000.)
    ax.plot(processes, t, 'k.-', label=k)
    ax.plot(processes, 1 / np.array(processes, dtype=float) * t[0], 'k:')
    ax.set_title(k)
    plt.xlabel('# processes')
    plt.ylabel('time per 1k windows (s)')

    ax = fig.add_subplot(2, len(signals), i + len(signals))
    ax.plot(results[k], 'k')
    ax.set_ylabel('reads\nper bin')
    ax.set_xlabel('bins')
    i += 1
fig.subplots_adjust(left=0.1, right=0.95, wspace=0.5, hspace=0.5)
fig.suptitle('%s features, %sbp each, chunksize=%s' % (args.nfeatures, size, args.chunksize))
fig.savefig(plot_filename)
plt.show()



"""
Module to handle custom colormaps.

`cmap_powerlaw_adjust`, `cmap_center_adjust`, and
`cmap_center_adjust` are from
https://sites.google.com/site/theodoregoetz/notes/matplotlib_colormapadjust
"""

import math
import copy
import numpy
import numpy as np
from matplotlib import pyplot, colors, cm
import matplotlib
import colorsys


def color_test(color):
    """
    Figure filled in with `color`; useful for troubleshooting or experimenting
    with colors
    """
    if isinstance(color, np.ndarray):
        color = color.ravel()
    fig = pyplot.figure(figsize=(2, 2), facecolor=color)


def smart_colormap(vmin, vmax, color_high='#b11902', hue_low=0.6):
    """
    Creates a "smart" colormap that is centered on zero, and accounts for
    asymmetrical vmin and vmax by matching saturation/value of high and low
    colors.

    It works by first creating a colormap from white to `color_high`.  Setting
    this color to the max(abs([vmin, vmax])), it then determines what the color
    of min(abs([vmin, vmax])) should be on that scale.  Then it shifts the
    color to the new hue `hue_low`, and finally creates a new colormap with the
    new hue-shifted as the low, `color_high` as the max, and centered on zero.

    :param color_high: a matplotlib color -- try "#b11902" for a nice red
    :param hue_low: float in [0, 1] -- try 0.6 for a nice blue
    :param vmin: lowest value in data you'll be plotting
    :param vmax: highest value in data you'll be plotting
    """
    # first go from white to color_high
    orig_cmap = matplotlib.colors.LinearSegmentedColormap.from_list(
        'test', ['#FFFFFF', color_high], N=2048)

    # For example, say vmin=-3 and vmax=9.  If vmin were positive, what would
    # its color be?
    vmin = float(vmin)
    vmax = float(vmax)
    mx = max([vmin, vmax])
    mn = min([vmin, vmax])
    frac = abs(mn / mx)
    rgb = orig_cmap(frac)[:-1]

    # Convert to HSV and shift the hue
    hsv = list(colorsys.rgb_to_hsv(*rgb))
    hsv[0] = hue_low
    new_rgb = colorsys.hsv_to_rgb(*hsv)
    new_hex = matplotlib.colors.rgb2hex(new_rgb)

    zeropoint = -vmin / (vmax - vmin)

    # Create a new colormap using the new hue-shifted color as the low end
    new_cmap = matplotlib.colors.LinearSegmentedColormap.from_list(
        'test', [(0, new_rgb), (zeropoint, '#FFFFFF'), (1, color_high)],
        N=2048)

    return new_cmap


def cmap_powerlaw_adjust(cmap, a):
    """
    Returns a new colormap based on the one given
    but adjusted via power-law, `newcmap = oldcmap**a`.

    :param cmap: colormap instance (e.g., cm.jet)
    :param a: power
    """
    if a < 0.:
        return cmap
    cdict = copy.copy(cmap._segmentdata)
    fn = lambda x: (x[0] ** a, x[1], x[2])
    for key in ('red', 'green', 'blue'):
        cdict[key] = map(fn, cdict[key])
        cdict[key].sort()
        assert (cdict[key][0] < 0 or cdict[key][-1] > 1), \
            "Resulting indices extend out of the [0, 1] segment."
    return colors.LinearSegmentedColormap('colormap', cdict, 1024)


def cmap_center_adjust(cmap, center_ratio):
    """
    Returns a new colormap based on the one given
    but adjusted so that the old center point higher
    (>0.5) or lower (<0.5)

    :param cmap: colormap instance (e.g., cm.jet)
    :param center_ratio:

    """
    if not (0. < center_ratio) & (center_ratio < 1.):
        return cmap
    a = math.log(center_ratio) / math.log(0.5)
    return cmap_powerlaw_adjust(cmap, a)


def cmap_center_point_adjust(cmap, range, center):
    """
    Converts center to a ratio between 0 and 1 of the
    range given and calls cmap_center_adjust(). returns
    a new adjusted colormap accordingly

    :param cmap: colormap instance
    :param range: Tuple of (min, max)
    :param center: New cmap center
    """
    if not ((range[0] < center) and (center < range[1])):
        return cmap
    return cmap_center_adjust(
        cmap,
        abs(center - range[0]) / abs(range[1] - range[0]))


if __name__ == '__main__':
    ### create some 2D histogram-type data
    def func3(x, y):
        return (1 - x / 2 + x ** 5 + y ** 3) * numpy.exp(-x ** 2 - y ** 2)
    x = numpy.linspace(-3.0, 3.0, 60)
    y = numpy.linspace(-3.0, 3.0, 60)
    X, Y = numpy.meshgrid(x, y)
    Z = func3(X, Y)
    extent = [x[0], x[-1], y[0], y[-1]]

    plotkwargs = {
        'extent': extent,
        'origin': 'lower',
        'interpolation': 'nearest',
        'aspect': 'auto'}

    fig = pyplot.figure(figsize=(8, 3))
    fig.subplots_adjust(left=.05, bottom=.11, right=.94, top=.83, wspace=.35)
    ax = [fig.add_subplot(1, 3, i) for i in range(1, 4, 1)]

    cmap = cm.seismic

    plt = ax[0].imshow(Z, cmap=cmap, **plotkwargs)
    cb = ax[0].figure.colorbar(plt, ax=ax[0])
    ax[0].set_title('cmap: seismic')

    plt = ax[1].imshow(Z, cmap=cmap_center_adjust(cmap, 0.75), **plotkwargs)
    cb = ax[1].figure.colorbar(plt, ax=ax[1])
    ax[1].set_title('center raised by 25%')

    plt = ax[2].imshow(
        Z,
        cmap=cmap_center_point_adjust(
            cmap, [numpy.min(Z), numpy.max(Z)], 0),
        **plotkwargs)
    cb = ax[2].figure.colorbar(plt, ax=ax[2])
    ax[2].set_title('center set to 0')

    pyplot.show()



"""
module for testing the larger files (x.bam, x.bed.gz, etc)
"""
import multiprocessing
import metaseq
import pybedtools

CPUS = multiprocessing.cpu_count()

gs = {}
for kind in ['bam', 'bigwig', 'bed', 'bigbed']:
    if kind == 'bed':
        ext = 'bed.gz'
    else:
        ext = kind
    gs[kind] = metaseq.genomic_signal(
        metaseq.example_filename('x.%s' % ext), kind)

# generate the test features
features = pybedtools.BedTool()\
        .window_maker(
            b=pybedtools.BedTool('chr2L 0 500000',
                                 from_string=True).fn,
            w=1000)\
        .shuffle(seed=1,
                 genome={'chr2L': (0, 5000000)})

args = (features,)
kwargs = dict(processes=CPUS, bins=100)
bam_array = gs['bam'].array(*args, **kwargs)
bed_array = gs['bed'].array(*args, **kwargs)
bw_array = gs['bigwig'].array(*args, method='get_as_array', **kwargs)
bb_array = gs['bigbed'].array(*args, **kwargs)


assert (bam_array == bed_array).all()
assert (bb_array == bed_array).all()
assert (bw_array == bed_array).all()

assert (bb_array == bam_array).all()
assert (bw_array == bam_array).all()

assert (bw_array == bb_array).all()



from metaseq import genomic_signal
import pybedtools
import numpy as np
try:
    del a
except NameError:
    pass
features = pybedtools.BedTool(pybedtools.BedTool('chr17.exons'))#.saveas()
gs = genomic_signal('data/wgEncodeHaibTfbsK562Atf3V0416101RawRep1_chr17.bigWig', 'bigwig')
a = gs.array(features, bins=100, processes=8, chunksize=250)



from metaseq.results_table import DESeqResults
from pybedtools.contrib.bigbed import bigbed

d = DESeqResults('data/rrp40-s2-polyA.final.summary', dbfn='data/dmel-all-r5.33-cleaned.gff.db')

d = d[:100]

x = d.colormapped_bedfile()
asql = d.autosql_file()

output = bigbed(x, 'dm3', 'test.bb', _as=asql, bedtype='bed9+5')
